import importlib as imp
import os,glob,json,numpy as np, pandas as pd, tifffile,mrcfile,scipy.optimize as opt
from typing import TYPE_CHECKING, Dict, Iterable, Optional, Sequence, Union
from subprocess import Popen,PIPE,check_output
from utils import glob_colors as colors
from crystals import Crystal,Lattice
from EDutils import utilities as ut         #;imp.reload(ut)
from multislice.rotating_crystal import get_crystal_rotation
from gemmi import cif
from . import import_ED as ED               ;imp.reload(ED)

class Dials(ED.Dataset):
    def __init__(self,path:str):
        '''Importing dials information
        It expects a .expt file and a reflections.txt file
        '''
        self.path = path
        refl_file = os.path.join(self.path,'reflections.txt')
        if not os.path.exists(refl_file):
            raise Exception('%s file not found' %refl_file)

        df = load_dials_reflections(refl_file)
        self.rpl = df[['h','k','l','I']].copy()
        self.rpl['hkl'] = [str((h,k,l)) for h,k,l in self.rpl[['h','k','l']].values]
        self.rpl[['qx','qy']] = df[['s1x','s1y']]
        self.rpl[['px','py']] = df[['o_px','o_py']]
        # self.rpl[['qy']]*=-1
        self.rpl['F'] = np.array(np.round(df['o_pz'])+1,dtype=int)

        #integrated intensities
        self.hkl = self.rpl[['I','hkl']].groupby('hkl').sum()
        self.hkl = self.hkl.drop(str((0,0,0)))

        self.rpl.index = self.rpl.hkl
        ######
        self.read_expt_file()
        self.init_geom()
        self.get_uvw()
        self.uvw0*=-1

        dyn_cif_file=os.path.join(self.path,'dials_dyn.cif_pets')
        if os.path.exists(dyn_cif_file):
            self.read_dyn_cif(dyn_cif_file)

    def xd_to_px(self,xd):
        '''
        - xd : Nx3 spot locations in the detector coordinate system
        '''
        px = xd[:,0]/xd[:,2]/self.dx  #;print('px:',px)
        py = xd[:,1]/xd[:,2]/self.dy  #;print('py:',py)
        return px,py

    def read_expt_file(self):
        expt_file=''
        for s in ['integrated','refined','indexed']:
            expt_file=os.path.join(self.path,'%s.expt' %s)
            if os.path.exists(expt_file):
                break
        try:
            with open(expt_file,'r') as f:
                expt = json.load(f)
        except:
            raise Exception('To import dials you at need a indexed.expt or refined.expt, integrated.expt ')

        print(colors.blue+'reading %s' %expt_file+colors.black)
        # expt = expt
        self.beam = -np.array(expt['beam'][0]['direction'])
        self.lam  = expt['beam'][0]['wavelength']
        detector  = expt["detector"][0]["panels"][0]
        self.orgx,self.orgy,self.F  = detector['origin']
        self.dx,self.dy             = detector['pixel_size']
        self.nxy                    = detector['image_size']
        oscillation = expt['scan'][0]['oscillation']
        self.hm     = expt['crystal'][0]['space_group_hall_symbol']

        #real space orientation matrix [a;b;c]
        self.A = np.array([
            expt['crystal'][0]['real_space_%s' %s] for s in 'abc'
            ])

        lat_c = np.linalg.norm(self.A,axis=1)
        angles = [ np.rad2deg(np.arccos(
                self.A[i].dot(self.A[j])/(lat_c[i]*lat_c[j])
            ))
            for i,j in zip([0,1,0],[1,2,2])]
        self.lat_params=[*lat_c,*angles]

        self.info={
            'ROTATION_AXIS'                 : expt['goniometer'][0]['rotation_axis'],
            'DATA_RANGE'                    : expt['scan'][0]['image_range'],
            'STARTING_ANGLE'                : oscillation[0],
            'OSCILLATION_RANGE'             : oscillation[1],
            'INCIDENT_BEAM_DIRECTION'       : self.beam,
            'DIRECTION_OF_DETECTOR_X-AXIS'  : detector['fast_axis'],
            'DIRECTION_OF_DETECTOR_Y-AXIS'  : detector['slow_axis'],
            'UNIT_CELL_CONSTANTS'           : self.lat_params,
        }
        e1,e2 = self.info['DIRECTION_OF_DETECTOR_X-AXIS'],self.info['DIRECTION_OF_DETECTOR_Y-AXIS']
        e3 = np.array([self.orgx,self.orgy,self.F])
        ED = np.array([e1,e2,e3]).T
        self.ED = ED


    def read_dyn_cif(self,dyn_cif_file):
        print(colors.blue+'reading %s' %dyn_cif_file+colors.black)
        d = load_dyn(dyn_cif_file)
        self.UB2 = d['UB']
        self.uvw = np.stack([d['u'],d['v'],d['w']]).T
        self.lat_params = d['lat_params']
        self.lat = np.array(Lattice.from_parameters(*self.lat_params).lattice_vectors)

        for k in ['alpha','n_frames'] : self.__dict__[k] = d[k]
        beams = self.lat.T.dot(self.uvw.T).T
        beams/=np.linalg.norm(beams,axis=1)[:,None]
        self.uvw0 = -beams

        # #pb dials duplicate first frame
        self.n_frames-=1
        self.alpha=self.alpha[1:]

        # self.uvw0=self.uvw0[1:]
        # print(np.cross(self.uvw0[0],self.uvw0[-1]))

        # predict_txt = os.path.join(self.path,'predict.txt')
        # if os.path.exists(predict_txt):
        #     self.df_pred = load_dials_predict(predict_txt)
        #
        # files = glob.glob(os.path.join(path,'**'),recursive=True)
        # imfile = ''
        # for ext in ['tiff','mrc']:
        #     images =[f for f in files if f.split('.')[-1]==ext]
        #     if any(images):
        #         imfile = images[0]
        #         print(colors.green+'image format %s detecte in %s' %(ext, imfile)+colors.black)
        # if imfile:
        #     if ext=='mrc':
        #         with mrcfile.open(imfile) as f:
        #             self.aper = f.extended_header["Pixel size X"][0]*1e-10 #A^-1
        #             self.nxy  = int(f.header.nx)
        #             self.Imax =  f.data.max()

def load_dials_predict(refl_txt):
    tmp = 'tmp.txt'
    cmd = "tail --lines=+25 %s | "\
        "sed -E 's/ +/ /g' | sed -E 's/^ //' | sed 's/,//g' "\
        "> %s" %(refl_txt, tmp)

    out = check_output(cmd,shell=True).decode()
    if out:print(out)
    df = pd.read_csv(tmp,
        names=[
            'h','k','l','id','panel','flag',
            'x','y','z',
            'px','py','pz',
            's1x','s1y','s1z',
            ],
        engine="python",sep=" ")

    out=check_output("rm %s" %tmp ,shell=True).decode()
    if out:print(out)
    df.index=[str(tuple(h)) for h in df[['h','k','l']].values]
    df['hkl']=df.index
    df['F'] = np.array(np.round(df['pz'])+1,dtype=int)
    return df

def load_dials_reflections(refl_txt):
    print(colors.blue+'reading %s' %refl_txt+colors.black)
    tmp = 'tmp.txt'
    cmd = "tail --lines=+39 %s | "\
        "sed -E 's/ +/ /g' | sed -E 's/^ //' | sed 's/,//g' "\
        "> %s" %(refl_txt, tmp)

    out = check_output(cmd,shell=True).decode()
    if out:print(out)
    df = pd.read_csv(tmp,
        names=[
            'h','k','l','id','i','panel','flag','I','sig',
            'c_qx','c_qy','c_qz','c_px','c_py','c_pz',
            'o_qx','o_qy','o_qz','vqx','vqy','vqz',
            'o_px','o_py','o_pz','vpx','vpy','vpz',
            's1x','s1y','s1z','Npix','rlpx','rlpy','rlpz'],
        engine="python",sep=" ")

    out=check_output("rm %s" %tmp ,shell=True).decode()
    if out:print(out)
    return df

def load_dyn_intensities(file_dyn):
    doc = cif.read_file(file_dyn )
    block = doc.sole_block()
    h = np.array([float(i) for i in list(block.find_loop('_refln_index_h'))])
    k = np.array([float(i) for i in list(block.find_loop('_refln_index_k'))])
    l = np.array([float(i) for i in list(block.find_loop('_refln_index_l'))])
    I = np.array([float(i) for i in list(block.find_loop('_refln_intensity_meas'))])
    sig = np.array([float(i) for i in list(block.find_loop('_refln_intensity_sigma'))])
    F = np.array([float(i) for i in list(block.find_loop('_refln_zone_axis_id'))])

    df = pd.DataFrame(columns=['h','k','l','I','sig','frame'])
    return df

def load_dyn(file_dyn):
    #  Now use dyn.cif_pets to set up the felix simulation input files
    doc = cif.read_file(file_dyn )
    block = doc.sole_block()

    # Extract data:
    # first the frames and their orientation for the felix.inp file
    # frame number fn
    fn = np.array([int(i) for i in list(block.find_loop('_diffrn_zone_axis_id'))])
    n_frames = len(fn)  # no. of frames
    # incident beam orientation u,v,w
    u = np.array([float(i) for i in list(block.find_loop('_diffrn_zone_axis_u'))])
    v = np.array([float(i) for i in list(block.find_loop('_diffrn_zone_axis_v'))])
    w = np.array([float(i) for i in list(block.find_loop('_diffrn_zone_axis_w'))])

    # Frame angles
    frame_alpha = np.array([float(i) for i in
                            list(block.find_loop('_diffrn_zone_axis_alpha'))])
    # Unit cell
    cell_a = float(block.find_value('_cell_length_a'))
    cell_b = float(block.find_value('_cell_length_b'))
    cell_c = float(block.find_value('_cell_length_c'))
    alpha = float(block.find_value('_cell_angle_alpha'))
    beta  = float(block.find_value('_cell_angle_beta'))
    gamma = float(block.find_value('_cell_angle_gamma'))
    UB = np.reshape(
        [float(block.find_value('_diffrn_orient_matrix_UB_%d%d' %(i,j)))
            for i,j in zip(*[x.flatten()
                for x in np.meshgrid(*(np.arange(3)+1,)*2) ])
        ],(3,3)).T
    cell=[cell_a,cell_b,cell_c]
    lat_params = [cell_a,cell_b,cell_c,alpha,beta,gamma]
    return dict(u=u,v=v,w=w,n_frames=n_frames,alpha=frame_alpha,cell=cell,lat_params=lat_params,UB=UB)



import numpy as np, pandas as pd
df_file = 'space_groups.pkl'
df = pd.read_pickle(df_file)

def apply_sym_cif(cif):
    with open(cif,'r') as f:lines=f.readlines()

    #### ops
    loops = [i for i,l in enumerate(lines) if 'loop_' in l]
    li = [i for i,l in enumerate(lines) if '_symmetry_equiv_pos_as_xyz' in l][0]
    ops = [l.split("'")[1] for l in lines[li+1:loops[-1]]]
    #### positions
    li = [i for i,l in enumerate(lines) if '_atom_site_occupancy' in l][0]
    lf = [i for i,l in enumerate(lines) if '#End' in l][0]
    xyz =np.array([np.array(l.split('  ')[1:4],dtype=float) for l in lines[li+1:lf]])

    X = []
    for x0 in xyz:
        x,y,z = x0
        X += [np.array(eval(op)) for op in ops]    #;print(X)
    X = np.array(X)
    X[X<0]+=1
    X[X>1]-=1
    X = np.unique(np.array(X),axis=0)
    return X

def apply_sym(xyz,no):
    '''
    xyz : asymmetric unit cell
    no : space group
    '''

    c=df.loc[no]
    spg = c['space group']                              #;print(spg)
    ops = [l for l in c.ops if l.strip() != '']         #;print(ops)
    X = np.array([np.array(eval(op)) for op in ops])    #;print(X)
    X[X<0]+=1
    X[X>1]-=1
    dx[i] = X
    return X

if __name__=="__main__":
    cif='/home/tarik/Documents/git/github/Felix/samples/GaAs_short/felix.cif'
    with open(cif,'r') as f:lines=[l.strip() for l in f.readlines()]
    #### ops
    loops = [i for i,l in enumerate(lines) if 'loop_' in l]
    li = [i for i,l in enumerate(lines) if '_symmetry_equiv_pos_as_xyz' in l][0]
    ops = [l.split("'")[1] for l in lines[li+1:loops[-1]]]
    #### positions
    li = [i for i,l in enumerate(lines) if '_atom_site_occupancy' in l][0]
    lf = [i for i,l in enumerate(lines) if '#End' in l][0]
    xyz =np.array([np.array(l.split('  ')[1:4],dtype=float) for l in lines[li+1:lf]])

    # apply_sym_cif(cif)



import importlib as imp
import pickle,matplotlib,os
import numpy as np, pandas as pd
import scipy.fftpack as fft
from scipy.integrate import nquad,trapz,quad
import utils.displayStandards as dsp
import utils.physicsConstants as cst
import utils.glob_colors as colors

class Multislice:
    '''python multislice
    #### Structure
    - pattern : 2D or 3D pattern (see multi2D or multi3D)
    #### Parameters
    - keV   : float - electron wavelength (keV)
    - tilt  : float 2-tuple : beam tilt (degrees)
    - dz    : float - slice thickness
    - nz    : int - number of slices along propagation axis
    - Nz    : int - number of unit cells along propagation axis (Nz takes preference over nz if set)
    - nx    : int or 2-tuple - used sampling if defined
    - Nx    : int or 2-tuple - supercell size
    #### TDS
    - TDS    : Use thermal diffuse scattering
    - nTDS   : nb configurations
    - wobble : wobble parameters (see Multislice._wobble for more info  )
    #### aliases
    - ndeg         : alias for nx
    - slice_thick  : alias for dz
    #### Save/Verbose options
    - opts : str - 'q'(save reciprocal space slices) 'x'(save real space slices)
    - v    : bool or str - verbose option
    - iZs  : int - info are saved every iZs slices
    - iZv  : int - info are displayed every iZv slices
    #### Display options
    - ppopt : TVPQXBZY (see Multislice.display for more info)
    #### Misc
    - copt : limit bandwidth on propagator option
    - sg   : change the sign of the propagator
    - eps  : scale the strength of the potential

    '''
    def __init__(self,
        pattern,
        keV=200,tilt=0,dz=1,slice_thick=None,nz=0,Nz=None,nx=2**10,Nx=1,
        TDS=False,nTDS=8,ndeg=None,wobble=0.05,
        copt=1,eps=1,sg=-1,
        iZs=1,s_opts='q',opts=None,iZv=1,v=1,
        ppopt='',name='./unknown',**kwargs):
        self.Mversion = 1.1
        self.pattern  = pattern
        self.keV  = keV
        self.dz   = self._get_thickness(dz,slice_thick)
        self.nx   = self._alias(nx,ndeg)
        self.Nx   = Nx
        self.Nz   = 0 # incremented during propagation
        self.nz   = 0 # incremented during propagation
        self.tilt = tilt
        #energy dependent
        self.lam  = cst.keV2lam(keV)
        self.sig  = cst.keV2sigma(keV)
        self.k0   = 1/self.lam
        self.eps  = eps
        self.copt = copt
        self.sg   = sg
        #TDS
        self.TDS    = TDS
        self.nTDS   = nTDS
        self.wobble = self._wobble(wobble)
        #Misc

        self._set_name(name)
        self._set_ns(v)
        nz     = self._get_nz(nz,Nz)
        s_opts = self._alias(s_opts,opts)
        self.run(nz,iZs,iZv,s_opts,v)
        if ppopt:
            self.display(ppopt,self.basename,**kwargs)


    ############################################################################
    #### parameters
    def _get_nz(self,nz,Nz):
        if Nz:nz=int(Nz*self.ns)
        return nz
    def _get_thickness(self,dz,slice_thick):
        if slice_thick:dz=slice_thick
        return dz
    def _alias(self,param,alias):
        if alias : param=alias
        return param
    def _wobble(self,wobble):
        if isinstance(wobble,int) or isinstance(wobble,float): wobble = [wobble]*5
        if isinstance(wobble,dict) :
            wobbles = [0]*5
            for k,v in wobble.items():wobbles[k]=v
            wobble = wobbles
        return np.array(wobble)

    ############################################################################
    #### Misc
    def _set_name(self,name):
        self.name = os.path.basename(name)
        self.path = os.path.realpath(os.path.dirname(name))+'/'
        self.fullname = self.path+self.name

    def _set_ns(self,v):
        self.ns = int(np.round(self.ez/self.dz))
        self.dz = self.ez/self.ns
        if v:
            print('Slice thickness and number of slices per cell')
            print('     lattice param z=%.2fA, dz=%.2fA, nzs=%d' %(self.ez,self.dz,self.ns))

    def update_z(self,nz):
        self.Nz += int(nz/self.ns)
        self.nz += nz
        self.z = np.hstack([self.z,np.arange(self.nz)*self.dz])
        self.update(nz)

    def save(self,file):
        with open(file,'wb') as out :
            pickle.dump(self, out, pickle.HIGHEST_PROTOCOL)
        print(colors.green+"object saved : \n"+colors.yellow+file+colors.black)

    ############################################################################
    #### Display
    def display(self,ppopt,name='',**kwargs):
        ''' ppopt options :
        T(Transmission)P(propagator)Q(Psi_q)X(Psi_x)B(beams)Z(Psi_xz)Y(Psi_qz)
        '''
        if 'T' in ppopt:self.Tz_show( name=name+'Tz.svg' ,**kwargs)
        if 'Z' in ppopt:self.Za_show( name=name+'Za.png' ,**kwargs)
        if 'P' in ppopt:self.Pq_show( name=name+'Pq.svg' ,**kwargs)
        if 'Q' in ppopt:self.Qz_show( name=name+'Qz.svg' ,**kwargs)
        if 'X' in ppopt:self.Xz_show( name=name+'Xz.svg' ,**kwargs)
        if 'B' in ppopt:self.Bz_show( name=name+'Bz.svg' ,**kwargs)
        if 'Z' in ppopt:self.Xxz_show(name=name+'Xxz.png',**kwargs)
        if 'Y' in ppopt:self.Qxz_show(name=name+'Xxz.png',**kwargs)


##################################################################
#### misc functions
##################################################################
def load(filename):
    '''load a saved object'''
    with open(filename,'rb') as f : multi = pickle.load(f)
    return multi

def tilts_show(tilts,mp2,iBs,iZs,**kwargs):
    ''' display beams for a sequence of tilted simulations
    tilts,mp2 : tilts array and multislice objects
    - iBs,iZs : beam and slice indices (integer allowed)
    '''
    bopt,zopt,nz=isinstance(iBs,int),isinstance(iZs,int),1
    if bopt:iBs=np.array([iBs])
    if zopt:iZs=np.array([iZs])
    if isinstance(iZs,slice):
        stop = (mp2[0].z.size+1)*(iZs.stop<0)+iZs.stop #;print(stop)
        nz = int(np.ceil((stop-iZs.start)/iZs.step))   #;print(nz)
    else:
        nz = np.array(iZs).size
    Itbz = np.zeros((tilts.size,iBs.size,nz))
    for i,t in enumerate(tilts):
        Itbz[i,:,:] = mp2[i].getB(iBs)[iZs,:].T

    if zopt and bopt:
        plts=[tilts,Itbz[:,0,iZs],'bo-']
        dsp.stddisp(plts,labs=[r'$\theta(deg)$','$I$'],**kwargs)
    if bopt and not zopt:
        csZ = dsp.getCs('Reds',nz)
        plts1 = [[tilts,Itbz[:,0,iZ],[csZ[iZ],'o-'],'%dnm' %(z0/10)] for iZ,z0 in enumerate(mp2[0].z[iZs])]
        dsp.stddisp(plts1,labs=[r'$\theta(deg)$','$I$'],**kwargs)
    if zopt and not bopt:
        cs = dsp.getCs('Spectral',iBs.size)
        # for iZ in iZs:Iz[:,iZ]/=Iz[:,iZ].max()
        plts2=[ [tilts,Itbz[:,i],[cs[i],'o-'],'iB=%d' %(iB)] for i,iB in enumerate(iBs)]
        dsp.stddisp(plts2,labs=[r'$\theta(deg)$','$I$'],**kwargs)



from utils import displayStandards as dsp
from utils import glob_colors as colors
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np,pandas as pd
import easygui,os
from . import config as cg
from . import lattice as lat

if plt.rcParams['keymap.save']:
    plt.rcParams['keymap.save']=[]

class Wallpaper_gui():
    def __init__(self,image,path='',pg='p1',npyfile=None,config=None,opt=0,
            params={'a':1},asym_verts=[],P1=[0,0]):
        self.path  = path
        self.name  = ''.join(os.path.basename(image).split('.')[:-1])#;print(self.name)
        self.image = image
        self.im0   = plt.imread(self.image)
        self._set_keys()
        self.pg  = pg
        if not self.load_config(config):
            self.params = params
            self.P1     = np.array(P1)
            self.rot    = 0
        self._set_plane_group()
        self.set_npy_file(npyfile)

        self.xylims=[]
        self.fig,self.ax = dsp.stddisp(title=self.pg)#opt='')
        nx,ny = self.im0.shape[:2]
        self.xylims=np.array([0,nx,0,ny])
        self.fig.canvas.mpl_connect('key_press_event',self)
        self.show()
        if opt:self.show_pattern(opt)

    ################################################################
    ### callbacks
    ################################################################
    def __call__(self,event):
        key = event.key         #;print(key)
        kopts=self.opts_keys
        if key in self.d_keys.keys():
            self.__getattribute__(self.d_keys[key])()
            self.show()

    def _set_keys(self):
        opts = ['vec','asym']
        self.opts=dict(zip(opts,[True]*len(opts)))
        self.opts_keys=dict(zip(['2','3'],opts))
        self.d_keys={
            'h':'help',
            'ctrl+enter'        :'set_vectors',
            'shift+enter'       :'set_origin',
            'shift+ctrl+enter'  :'set_plane_group',
            'ctrl+2'            :'show_pattern',
            'ctrl+3'            :'show_unit_cell',
            'ctrl+4'            :'show_wallpp',
            'ctrl+='            :'increase_lims',
            'ctrl+-'            :'decrease_lims',
            'ctrl+o'            :'open_image',
            'ctrl+O'            :'load_config',
            'ctrl+S'            :'save_config',
            'ctrl+s'            :'save',
            'S'                 :'save_pattern',
            's'                 :'save_fig'
        }
        self.df_keys = pd.DataFrame.from_dict({v:k for k,v in self.d_keys.items()},orient='index',columns=['key'])

    def increase_lims(self):self.xylims *= 2
    def decrease_lims(self):self.xylims = self.xylims/2

    def set_vectors(self):
        print('select points defining lattice parameters')
        ns = {'oblique':3,'rect':3,'hex':2,'square':2}
        lat_type=self.lat_type
        n = ns[lat_type]
        pts = np.array(plt.ginput(n=n,timeout=1000))
        if len(pts)==n:
            if lat_type=='square' or lat_type=='hex':
                a1 = pts[1]-pts[0]
                a = np.linalg.norm(a1)
                params = {'a':int(a)}
            else:
                P1,P2,P3 = pts
                a1,a2 = P2-P1,P3-P1
                a,b = np.linalg.norm(a1),np.linalg.norm(a2)
                params = {'a':int(a),'b':int(b)}
                if lat_type=='oblique':
                    alpha = np.arccos(np.dot(a1,a2)/(a*b))
                    params['alpha'] = np.rad2deg(alpha)
            rot = np.rad2deg(np.arctan2(a1[1],a1[0]))#*np.sign(a1[1])
            eps = rot-90*np.round(rot/90)           #;print(rot,eps)
            if abs(eps)<8:rot=90*np.round(rot/90)   #;print(rot)
            self.rot = -np.deg2rad(rot)
            self.params = params
            self._set_plane_group()

    def set_origin(self):
        print('select an origin ')
        pts = np.array(plt.ginput(n=1,timeout=10000),dtype=int)
        self.P1=pts[0]

    def set_plane_group(self):
        pg = easygui.choicebox('choose plane group','plane',cg.pp_types)
        if pg :
            self.pg=pg
            self._set_plane_group()
            self.set_npy_file()

    def _set_plane_group(self):
        self.lat_type = cg.df_wallpp.loc[self.pg,'lat_type']
        asym_cell_frac = cg.df_wallpp.loc[self.pg,'asym_cell']
        self.lat_vec = lat.set_lattice_vec(self.lat_type,**self.params)
        self.asym_verts = asym_cell_frac.dot(self.lat_vec)
        if self.rot:
            ct,st = np.cos(self.rot),np.sin(self.rot)
            rot2=np.array([[ct,st],[-st,ct]])
            self.lat_vec = rot2.dot(self.lat_vec.T).T
            self.asym_verts = rot2.dot(self.asym_verts.T).T

    def set_pattern(self):
        nx = min(self.im0.shape[:2])
        patch = plt.Polygon(self.asym_verts+self.P1)
        x,y = np.meshgrid(np.arange(nx),np.arange(nx))
        x0,y0 = x.flatten(),y.flatten()
        # print(patch.contains_point([320,90]))
        # im0 = np.sum(self.im0[:nx,:nx,:],axis=2)
        im0 =  self.im0[:nx,:nx,:]#/255#-self.im0[:nx,:nx,0]
        idx = patch.contains_points(list(np.array([x0,y0]).T),radius=0)#;print(idx)
        x0,y0,z0,z1,z2 = x0[idx],y0[idx],im0[:,:,0].flatten()[idx],im0[:,:,1].flatten()[idx],im0[:,:,2].flatten()[idx]
        # z = z1
        # u = np.linalg.norm(np.vstack([x0,y0]).T-np.array([320,90]),axis=1)
        # print(x0[u.argmin()],y0[u.argmin()])

        self.vec = self.lat_vec
        x0-=self.P1[0]
        y0-=self.P1[1]
        if self.rot:
            ct,st = np.cos(-self.rot),np.sin(-self.rot)
            rot2  = np.array([[ct,st],[-st,ct]])
            x0,y0 = rot2.dot(np.vstack([x0,y0]))
            self.vec = rot2.dot(self.vec.T).T
        self.pattern = np.array([x0,y0,z0,z1,z2]).T

    ################################################################
    ### show
    ################################################################
    def show(self):
        self.ax.cla()
        self.ax.imshow(self.im0,origin='lower')
        pp = []
        if self.opts['vec']:
            a1,a2 = self.lat_vec
            verts = self.P1+np.array([0*a1,a1,a1+a2,a2])
            pp += [patches.Polygon(verts,alpha=0.2,color='b')]
        if self.opts['asym']:
            verts = self.P1+self.asym_verts
            pp += [patches.Polygon(verts,alpha=0.2,color='r')]

        dsp.stddisp(ax=self.ax,patches=pp,pOpt='tX',xylims=self.xylims,opt='',title=self.pg)
        self.fig.canvas.draw()

    def show_asym(self)     :self.show_pattern(0)
    def show_unit_cell(self):self.show_pattern(1)
    def show_wallpp(self)   :self.show_pattern(2)
    def show_pattern(self,opt=0):
        npyfile=os.path.join(self.path,'tmp.npy')
        self.save(npyfile)
        pattern,vec = np.load(npyfile,allow_pickle=True)
        a1,a2 = vec
        verts = [0*a1,a1,a1+a2,a2]
        pp = [patches.Polygon(verts,alpha=0.5,color='b')]
        # pattern = np.array(pattern).T
        if opt:pattern = cg.generate(self.pg,vec,pattern)
        if opt>1:
            pattern=cg.replicate(2,2,self.pg,self.params,pattern)
        x,y = pattern[:,:2].T
        F = pattern[:,2:] #;print(F.shape)
        dsp.stddisp(scat=[x,y,F],sargs={'s':40,'marker':'s'},patches=pp,
            title=self.pg,equal=1)

    ################################################################
    ### help,save,load,open
    ################################################################
    def help(self):
        print(colors.green+'help:')
        print(colors.blue,self.df_keys)
        print(colors.green+'good luck'+colors.black)

    def save_config(self,config=None):
        if not config:config=os.path.join(self.path,self.name+'%s_config.npy' %self.pg)
        np.save(config,[self.pg,self.params,self.rot,self.P1])
        print(colors.green+'config saved : '+colors.yellow+config+colors.black)

    def load_config(self,config=None):
        if config:
            if not isinstance(config,str):config=os.path.join(self.path,self.name+'%s_config.npy' %self.pg)
            if os.path.exists(config):
                self.pg,self.params,self.rot,self.P1 = np.load(config,allow_pickle=True)
                self._set_plane_group()
                print(colors.green+'config loaded : '+colors.yellow+config+colors.black)
                return 1

    def open_image(self):
        image = easygui.fileopenbox('choose image','image',self.image)
        if image:
            self.image=image
            self.im0 = plt.imread(self.image)
            self.set_npy_file()

    def set_npy_file(self,npyfile=None):
        if not npyfile:
            self.npyfile = os.path.join(self.path,self.name+'_%s.npy' %self.pg)
        else:
            self.npyfile=npyfile

    def save_pattern(self):
        npyfile = easygui.filesavebox('save pattern','save',self.npyfile)
        self.set_npy_file(npyfile)
        if npyfile:self.save()

    def save(self,npyfile=''):
        if not npyfile:npyfile = self.npyfile
        self.set_pattern()
        np.save(npyfile,[self.pattern,self.vec])
        print(colors.green+'file saved : '+colors.yellow+npyfile+colors.black)
    def save_fig(self):
        figname = os.path.join(self.path,self.name+'.png')
        dsp.saveFig(figname,ax=self.ax)



import importlib as imp
import pickle,matplotlib,os
import numpy as np, pandas as pd
import scipy.fftpack as fft
# from scipy.integrate import nquad,trapz,quad
import utils.displayStandards as dsp
import utils.physicsConstants as cst
import utils.glob_colors as colors
from . import postprocess as pp     ;imp.reload(pp)
from . import pymultislice          ;imp.reload(pymultislice)

atoms = pd.DataFrame.from_dict(
    { 1:[colors.unicolor(0.9),40],
      6:[colors.unicolor(0.5),60],7:[(0,0,1),70], 8:[(1,0,0),80],
      14:[(1,0.75,0.5),100],16:[(0,1,1),100]},
    orient='index',columns=['c','s'])
Ai = {1:1.0, 6:1.5, 7:1.5, 8:1.5, 14:5.0, 16:2.0}
Zs = {1:0, 6:1, 7:2, 8:3, 14:4, 16:5}
fsplines = os.path.dirname(__file__)+'/data/splines.npy'
rx2,y_vz,b_vz,c_vz,d_vz = np.load(fsplines,allow_pickle=True)
def fv(z,y,x,za,ya,xa,Za):
    r2 = (x-xa)**2+(y-ya)**2#+(z-za)**2
    # V = np.exp(-r2/Ai[Za])
    # return  np.exp(-r2/Ai[Za])
    iz = Zs[Za]
    i  = np.argmin(abs(r2-rx2))
    z  = r2 - rx2[i];
    vz = y_vz[iz][i] + ( b_vz[iz][i] + ( c_vz[iz][i] + d_vz[iz][i] *z ) *z) *z;
    return vz
# fv = lambda z,y,x,za,ya,xa,Za:np.exp(-((np.sqrt(x**2+y**2+z**2)-np.sqrt(xa**2+ya**2+za**2))/Ai[Za]) **2)



class Multi3D(pymultislice.Multislice):
    '''multislice 3D for quick testing
    - pattern  : list of 4d-arrays - [Z,x,y,z]
    - ax,by,cz : lattice constants
    #### beam recording
    - hk    : int or 2-list of list - number/indices of beams to record
    - hkopt : str - 's'(sym) 'r'(repeat) 'o'(include origin)
    #### aliases
    - nxy : alias for nx
    - Nxy : alias for Nx
    #### other
    - **kwargs : see pymultislice.Multislice.__init__
    - s_opts : 's'(save object) 'x'(real space), 'q'(reciprocal space) 't'(transmission function)
    - temsim : bool - same implemetation as temsim if True
    Example :
    Multi3D(pattern,ax,by,cz,
        keV=200,Nx=1,tilt=0,dz=1,nz=1,
        TDS=False,nTDS=8,nx=2**10,wobble=0.05,
        copt=1,eps=1,sg=-1,
        iZs=1,opts='q',iZv=1,v=1,ppopt='')
    '''
    def __init__(self,pattern,ax,by,cz,
        nxy=None,Nxy=None,
        hk=None,hkopt='sr',temsim=True,
        **kwargs):
        self.version = 1.0
        self.ax  = ax
        self.by  = by
        self.cz  = cz
        self.ez  = self.cz
        self.nxy = nxy
        self.Nxy = Nxy
        self.h,self.k = self._set_beams(hk,hkopt)
        self.temsim   = temsim
        print(colors.red+'\t\t 3D multislice simulation '+colors.black)
        super().__init__(pattern,**kwargs)

    def _init_3D_params(self):
        self.nxy  = self._alias3D(self.nx,self.nxy)
        self.Nxy  = self._alias3D(self.Nx,self.Nxy)
        self.tilt = self._alias3D(self.tilt,None)
        self.nx,self.ny     = self.nxy
        self.Nx,self.Ny     = self.Nxy
        self.axby           = np.array([self.ax,self.by])
        self.dx,self.dy     = np.array([self.ax,self.by])*self.Nxy/self.nxy
        self.dqx,self.dqy   = 1/np.array([self.ax,self.by])*self.Nxy
        self.dxy  = np.array([self.dx,self.dy])
        self.dqxy = np.array([self.dqx,self.dqy])
        ## bandwidth limit
        self.q2max = (2/3*min(self.nxy/(2*self.axby)))**2 #prevent aliasing

        self.Psi_x = np.ones(self.nxy)#/np.sqrt(self.nx*self.ny*self.dx*self.dy) #;print((self.Psi_x**2).sum()*self.dx*self.dy)
        self.iz = 0
        self.Nhk   = self.h.size
        self.beams = np.zeros((self.Nhk,0),dtype=complex)
        self.hk = [(h0,k0) for h0,k0 in zip(self.h,self.k)]
        self.z = np.array([])

    def update(self,nz):
        self.beams = np.hstack([ self.beams,np.zeros((self.Nhk,nz),dtype=complex) ])
    def _get_z(self,iz=None):
        if not isinstance(iz,int):
            z = self.z.copy()
        else:
            z = self.z[iz]
        if self.temsim:
            z+= 0.75*self.dz
        return z

    def _alias3D(self,param,alias):
        if alias : param=alias
        if isinstance(param,int):param=[param]*2
        return np.array(param)

    def _set_beams(self,hk,hkopt):
        if isinstance(hk,int):hk=[hk]*2
        if isinstance(hk,list):
            Nh,Nk = hk
            nh,nk = 1,1
            if 'r' in hkopt : nh,nk = self.Nxy
            h,k = np.meshgrid(np.arange(0,Nh+1)*nh,np.arange(0,Nk+1)*nk)
            h,k = h.flatten(),k.flatten()
            if 'o' not in hkopt:
                h,k = h[1:],k[1:]
            if 's' in hkopt:
                hsym,ksym = np.meshgrid(np.arange(-Nh,1)*nh,np.arange(-Nk,1)*nk)
                hsym,ksym = hsym[:-1],ksym[:-1]
                h,k = np.hstack([h,hsym.flatten()]),np.hstack([k,ksym.flatten()])
        return h,k

    def save(self):
        super().save(self.fullname+'_3D.pkl')

    ##################################################################
    ###### Display
    ##################################################################
    def Vz_show(self,iz,**kwargs):
        print(colors.blue+'...Integrating projected potential...'+colors.black)
        Vz = self._projected_potential(iz)
        nx,ny = np.array(self.nxy/self.Nxy,dtype=int)
        x,y   = np.meshgrid(np.arange(nx)*self.dx, np.arange(ny)*self.dy)  #;print(x.shape,y.shape,Vz.shape)
        im    = [x,y,Vz.T]
        Za,xa,ya  = self.pattern[self.Zas[iz]:self.Zas[iz+1],:3].T
        Zatom   = atoms.loc[Za]
        scat    = [xa,ya,Zatom.s,Zatom.c]

        tle = '$V_z$ for slice z=%.2fA, iz=%d' %(self._get_z(iz),iz)
        dsp.stddisp(im=im,scat=scat,labs=['$x$','$y$'],title=tle,
            caxis=[0,Vz.max()],imOpt='c',cs='I',axPos='V',pOpt='pt',#gridOn=False,
            **kwargs)

    def Tz_show(self,iz,opts='i',save=0,load=1,**kwargs):
        '''Plot the transmission function for slice iz :
        - iz : int - slice number
        - opts : str - 'i'(show im(T)) 'r'(show Re(T))
        '''
        T = self._get_transmission_function(iz,1,self.copt,save,load).T
        nx,ny = np.array(self.nxy/self.Nxy,dtype=int)
        x,y   = np.meshgrid(np.arange(nx)*self.dx, np.arange(ny)*self.dy)  #;print(x.shape,y.shape,Vz.shape)
        if 'i' in opts:Tz = T.imag
        if 'r' in opts:Tz = T.real
        im    = [x,y,Tz]

        Za,xa,ya  = self.pattern[self.Zas[iz]:self.Zas[iz+1],:3].T
        Zatom   = atoms.loc[Za]
        scat    = [xa,ya,Zatom.s,Zatom.c]

        tle = ' $%s(T)$ for slice z=%.2fA, iz=%d' %(['Re','Im']['i' in opts],self._get_z(iz),iz)
        dsp.stddisp(im=im,scat=scat,labs=['$x$','$y$'],title=tle,
            caxis=[0,Tz.max()],pOpt='im',
            **kwargs)

    def Pq_show(self,opts='qri',**kwargs):
        P = self.Pq
        Px,Py = P[0,:],P[:,0]
        if 'q' in opts:
            x = np.arange(self.nx)
            plts = [[x,Px.real,'b','Re Px'],[x,Px.imag,'r','Im Px']]
            plts+= [[x,Py.real,'c','Re Py'],[x,Py.imag,'m','Im Py']]
            dsp.stddisp(plts,labs=['q','Px,Py'])

        if 'r' in opts:dsp.stddisp(im=[P.real],pOpt='im',title='re P',**kwargs)
        if 'i' in opts:dsp.stddisp(im=[P.imag],pOpt='im',title='im P',**kwargs)
        return Px,Py

    def Bz_show(self,iBs=slice(1,None,None),**kwargs):
        if isinstance(iBs,list):
            if isinstance(iBs[0],tuple):
                iBs = [i for i,hk0 in enumerate(self.hk) if hk0 in iBs]

        beams = self.beams[iBs,:]
        hk      = np.array(['(%d_%d,%d_%d)' %(int(h0/self.Nx),h0%self.Nx,int(k0/self.Ny),k0%self.Ny ) for h0,k0 in zip(self.h,self.k)])[iBs]
        t       = self._get_z()
        re,im   = np.real(beams),np.imag(beams)
        Ib      = np.abs(beams)**2
        beams   = [hk,t,re,im,Ib]
        return pp.plot_beam_thickness(beams,**kwargs)



    ####################################################################################################################################
    ###### main computations
    ####################################################################################################################################
    def run(self,nz,iZs,iZv,s_opts,v):
        self._init_3D_params()
        self._propagator(v)
        self._sort_atoms(v)
        self.propagate(nz,iZs,iZv,s_opts,v)
        if 's' in s_opts:self.save()

    def _propagator(self,v=1):
        if v:print(colors.blue+'...Setting propagator...'+colors.black)
        sg,copt = self.sg,self.copt
        qx,qy = np.meshgrid(fft.fftfreq(self.nx,self.dx), fft.fftfreq(self.ny,self.dy))
        q2 = qx**2+qy**2
        # if self.tilt: #isinstance(tilt,np.ndarray) or isinstance(tilt,list):
        #     kx = self.k0*np.sin(self.tilt*np.pi/180)
        #     self.Pq = np.exp(sg*1J*np.pi*self.dz*(self.q+kx)**2/self.k0)
        # else:
        self.Pq = np.exp(self.sg*1J*np.pi*self.dz*q2/self.k0)
        self.bw_mask = q2>self.q2max
        if copt:self.Pq[self.bw_mask] = 0

    def _sort_atoms(self,v=1):
        if v:print(colors.blue+'...Sorting atoms per slice...'+colors.black)
        self.pattern = self.pattern[np.argsort(self.pattern[:,3]),:]
        #replicate one time along z
        pattern2 = self.pattern.copy(); pattern2[:,3]   += self.ez
        self.pattern = np.vstack([self.pattern,pattern2])
        za  = self.pattern[:,3]
        dz  = self.dz
        if self.temsim:
            if v>1:print('TEMSIM sorting ')
            self.zis = np.arange(-0.25*dz,self.ez+dz,dz)
            self.Zas = np.cumsum(np.histogram(za,self.zis)[0]) #;print(self.Zas)
            self.Zas = np.hstack([0,self.Zas])
        else:
            self.zis = np.arange(-dz,self.ez+1.1*dz,dz)
            self.Zas = np.cumsum(np.histogram(za,self.zis)[0]) #;print(self.Zas)

    def _projected_potential(self,iz):
        dx,dy,dz,Zas = self.dx,self.dy,self.dz,self.Zas
        nx,ny = np.array(self.nxy/self.Nxy,dtype=int) #unit cell nx,ny
        zi = iz*self.dz
        Vz = np.zeros((nx,ny))
        # print(iz,Zas,Zas.shape)
        for ia in range(Zas[iz],Zas[iz+1]):
            # print(ia)
            Za,xa,ya,za = self.pattern[ia,:4]
            Za = int(Za)
            nax,nay = int(Ai[Za]/dx),int(Ai[Za]/dy)               #;print('nax,nay',nax,nay)
            ixa,iya = int(xa/dx),int(ya/dy)
            for ix in ixa+np.arange(-nax,nax+1):
                x = dx*ix
                for iy in iya+np.arange(-nay,nay+1):
                    y = dy*iy
                    Vz[ix%nx,iy%ny] += fv(zi+dz/2,y,x,za,ya,xa,Za)
        return Vz

    def _transmission_function(self,iz=None,v=0,copt=1,save=0,load=0):
        if v:print(colors.blue+'...Integrating projected potential...'+colors.black)
        Vz = self._projected_potential(iz)
        Tz = np.exp(1J*self.sig*self.eps*Vz*1e-3)
        T  = np.zeros(self.nxy,dtype=complex)
        nx,ny = Tz.shape*self.Nxy
        T[:nx,:ny] = np.tile(Tz,self.Nxy)
        if copt:               #prevent aliasing
            if v>1:print(colors.blue+'...bandwidth limit transmission...'+colors.black)
            T = fft.fft2(T)
            T[self.bw_mask] = 0
            T = fft.ifft2(T)
        if save:
            filename =self.fullname+'_T%s.npy' %str(iz).zfill(3)
            np.save(filename,T)
            print(colors.green+'transmission saved : '+colors.yellow+filename+colors.black)
        return T

    def _get_transmission_function(self,iz=None,v=0,copt=1,save=0,load=0):
        if iz>self.ns or load:
            izl = iz%self.ns
            if not izl and iz>0:izl=self.ns
            filename =self.fullname+'_T%s.npy' %str(izl).zfill(3)
            try:
                if v>1:print(colors.green+'iz=%d,is=%d, loading ' %(iz,iz%self.ns)+colors.yellow+filename+colors.black)
                T = np.load(filename)
            except:
                T = self._transmission_function(iz,v,copt,save,load)
        else:
            T = self._transmission_function(iz,v,copt,save,load)
        return T

    def propagate(self,nz,iZs,iZv,s_opts,v):
        self.update_z(nz)
        if v:print(colors.green+'...Starting propagation loop...'+colors.black)
        save = 't' in s_opts
        for i in range(nz):
            T = self._get_transmission_function(self.iz,v,self.copt,save,'l' in s_opts)
            if v>1:print(colors.blue+'...FFT...'+colors.black)
            self.Psi_q = fft.fft2(T*self.Psi_x) #periodic assumption
            self.Psi_x = fft.ifft2(self.Pq*self.Psi_q)

            msg=''
            if v and (not i%iZv or i==nz-1):
                Ix2 = np.sum(np.abs(self.Psi_x)**2)/(self.nx*self.ny) #*self.dx*self.dy
                Iq2 = np.sum(np.abs(self.Psi_q/(self.nx*self.ny))**2)#/(self.dqx*self.dqy) #parseval's theorem of the DFT
                msg+='i=%-4d,z=%-7.3f A, I=%.4f, Iq=%.4f ' %(i,self._get_z(self.iz),Ix2,Iq2)

            # if not i%iZs and s_opts :
            self.beams[:,self.iz] = self.Psi_q[self.h,self.k]/(self.nx*self.ny)
                # if 'T' in s_opts:
                #     np.save('Tz%d.npy' %i,T)
                # if 'x' in s_opts:
                #     np.save('psi2%d_x.npy' %i,np.abs(self.Psi_x)**2)
                # if 'q' in s_opts:
                #     np.save('psi2%d_q.npy' %i,np.abs(self.Psi_q)**2)
            if msg:print(colors.blue+msg+colors.black)
            self.iz+=1


if __name__=='__main__':
    print('run tests/multislice/base_3D.py to see an example')



# import importlib as imp
import os,glob,json,numpy as np, pandas as pd, tifffile,mrcfile,scipy.optimize as opt
from crystals import Crystal,Lattice
from utils import physicsConstants as cst
from . import utilities as ut               #;imp.reload(ut)


class Dataset:
    def R(self,pz) :
        '''Rotation matrix at frame pz
        Parameters:
            pz : frame number
        '''
        a0 = self.info['STARTING_ANGLE']
        da = self.info['OSCILLATION_RANGE']
        f0 = self.info['DATA_RANGE'][0]
        angle = a0 + da*(pz-f0)
        R = ut.rotation_matrix(
            self.info['ROTATION_AXIS'],
            angle,
            deg=True)
        return R

    def init_geom(self):

        self.aper = 1/abs(self.lam*self.F/self.dx) #;print('aper : %.4f recAngstrom' %self.aper)
        self.keV = cst.lam2keV(self.lam)

        lat = Lattice.from_parameters(*self.lat_params)
        self.Alat=np.array(lat.lattice_vectors).T
        self.Arec=np.array(lat.reciprocal_vectors).T/(2*np.pi)
        a,b,c = self.A
        a_star = np.cross(b,c)/a.dot(np.cross(b,c))
        b_star = np.cross(c,a)/b.dot(np.cross(c,a))
        c_star = np.cross(a,b)/c.dot(np.cross(a,b))
        self.UB = np.array([a_star,b_star,c_star]).T

    def get_uvw(self):
        '''(u,v,w) are the components of the orientation vector in reciprocal cartesian space
        It is such that K-g as it appears in the blochwave excitation error
        can be computed where the components of the beam g are simply calculated
        from the crystal reciprocal lattice vectors and the beam miller indices (h,k,l)
        '''
        self.frames=np.arange(self.info['DATA_RANGE'][0],self.info['DATA_RANGE'][1])
        # mat=self.UB2
        beam = self.info['INCIDENT_BEAM_DIRECTION']
        self.uvw0 = np.array([
            self.Arec.dot(np.linalg.inv(self.R(f).dot(self.UB))).dot(beam)
                for f in self.frames])
        self.n_frames = self.uvw0.shape[0]
        self.cen = pd.DataFrame([[self.orgx,self.orgy]]*self.n_frames,
            columns=['px','py'])

    def hkl_to_pixels(self,h,frame):
        ''' convert miller indices to pixel locations
            - h : list of miller indices
            - frame : frame location of the miller indices
        '''
        # pz = self.rpl.loc[hkl,'pz'] #frame location of reflection
        #hlist of reflections with shape :n_refls x 3
        hkl = np.array([eval(hkl) for hkl in h])

        #### #reflection in reciprocal lab frame
        r = self.R(frame).dot(self.UB).dot(hkl.T).T
        #incident beam in reciprocal lab frame
        s0 = self.info['INCIDENT_BEAM_DIRECTION']/self.lam
        # scattering vector in reciprocal lab frame
        s = r+s0                                    #;print('s : ',s)

        #### reflection in reciprocal Orthonormal crystal frame
        # r = self.Arec.dot(hkl.T).T
        # #Incident Beam vector in reciprocal Orthonormal crystal frame
        # M = self.Arec.dot(np.linalg.inv(self.R(frame).dot(self.UB)))
        # u = M.dot(self.info['INCIDENT_BEAM_DIRECTION'])
        # s0 = u/(self.lam*np.linalg.norm(u))
        # #scattering vector in reciprocal Orthonormal crystal frame
        # s1 = r+s0
        # #scattering vector in reciprocal lab frame
        # s = (np.linalg.inv(M).dot(s1.T)).T          #;print('s1 : ',s1)


        #scattering vector direction in lab frame
        norms = np.linalg.norm(s,axis=1)            #; print('norms:',norms)
        s=(s.T/np.linalg.norm(s,axis=1)).T          #; print('s:\n',s);print('norms:',np.linalg.norm(s,axis=1))

        #location of spot in real space lab frame
        x = np.array([
            np.tan(np.arctan2(s[:,0],s[:,2]))*self.F,
            np.tan(np.arctan2(s[:,1],s[:,2]))*self.F,
            self.F*np.ones(s[:,1].shape)]).T
        #print('x:\n',x)
        # spot location in detector frame
        xd = np.linalg.inv(self.ED).dot(x.T).T          #;print('xd:',xd)
        px,py = self.xd_to_px(xd)

        #package
        df_pxy = pd.DataFrame()
        df_pxy['px'] = px
        df_pxy['py'] = py
        df_pxy.index = h
        # pixel locations

        return df_pxy


    def to_shelx(self,hkl,file='',output_dir=None):
        '''converts to a .hkl file ready to use by shelx
        hkl:
        '''
        if not isinstance(output_dir,str):output_dir=self.path
        if not file:file='refl.hkl'
        file_path = os.path.join(output_dir, file)
        ut.to_shelx(hkl,file_path)



import importlib as imp
import numpy as np
import pandas as pd
import os,glob
from utils import displayStandards as dsp
from utils import plt
from . import dyn_utils as dyn
from . import utilities as ut;imp.reload(ut)
from . import pets as pt;imp.reload(pt)

def load_felix(path):
    return ut.load_pkl(os.path.join(path,'felix.pkl'))

class Felix():
    '''Felix class for generating full simulations

    Parameters :
    ------------
        path
            simulation folder (must contain reflprofiles_strong.dat and *dyn.cif_pets )
        xtal
            crystal name
        ww
            width of the extracted rocking curve +/-
        keV
            wavelength
        frames_per_sim
            frames per simulation
        wid
            radius of the LACBED pattern in pixels (the patterns are 2*wid x 2*wid pixels)
    '''
    def __init__(self,path,xtal,
            ww=40,keV=200,frames_per_sim=20,wid=200):
        self.path = os.path.realpath(path)
        self.xtal = xtal
        self.ww   = ww
        self.wid  = wid
        self.nwx  = 2*self.wid
        self.file_profiles = os.path.join(self.path,'reflprofiles_strong.dat')
        self.file_dyn = glob.glob(os.path.join(self.path,'*dyn.cif_pets'))[0]
        self.paths = {
            'rock_exp' : os.path.join(self.path,'rock_curves'),
            'optim'    : os.path.join(self.path,'optimization'),
            'ref'      : os.path.join(self.path,'DWF_refinement'),
            'rock'     : os.path.join(self.path,'plots'),
            }
        for f in self.paths:
            if not os.path.exists(f):os.mkdir(f)
        self.import_rocks()
        print('... load pets info ... ')
        self.load_dyn()
        print('... integrating exp reflections ... ')
        self.integrate_exp_intensity()
        self.save()

    def save(self):
        ut.save_pkl(self,file=os.path.join(self.path,'felix.pkl'))

    def load_dyn(self):
        #  Now use dyn.cif_pets to set up the felix simulation input files
        dyn = pt.load_dyn(self.file_dyn)
        # print(dyn)
        self.__dict__.update(dyn)
        # self.u,self.v,self.w,self.n_frames,self.frame_alpha,self.cell =

        # doc = cif.read_file(self.file_dyn)
        # block = doc.sole_block()
        #
        # # Extract data:
        # # first the frames and their orientation for the felix.inp file
        # # frame number fn
        # fn = np.array([int(i) for i in list(block.find_loop('_diffrn_zone_axis_id'))])
        # self.n_frames = len(fn)  # no. of frames
        # # incident beam orientation u,v,w
        # self.u = np.array([float(i) for i in list(block.find_loop('_diffrn_zone_axis_u'))])
        # self.v = np.array([float(i) for i in list(block.find_loop('_diffrn_zone_axis_v'))])
        # self.w = np.array([float(i) for i in list(block.find_loop('_diffrn_zone_axis_w'))])
        #
        # # Frame angles
        # self.frame_alpha = np.array([float(i) for i in
        #                         list(block.find_loop('_diffrn_zone_axis_alpha'))])
        # # Unit cell
        # cell_a = float(block.find_value('_cell_length_a'))
        # cell_b = float(block.find_value('_cell_length_b'))
        # cell_c = float(block.find_value('_cell_length_c'))
        # self.cell=[cell_a,cell_b,cell_c]

    def load_refl_df(self):
        cols = list(self.df_refl.columns.values)
        with open (self.file_profiles,'r') as f:lines = f.readlines()
        #fix touching columns
        for i in range(len(lines)) : lines[i] = lines[i][:96]+' '+lines[i][96:]
        # get reflection block lines
        l_refl = [i for i,l in enumerate(lines) if '#' in l]+[len(lines)]
        for l1,l2 in zip(l_refl[:-1],l_refl[1:]):
            with open ('tmp.csv','w') as f:f.write(''.join(lines[l1+1:l2-2]))
            df=pd.read_csv('tmp.csv',sep=' +',engine='python',
                names=['h','k','l']+cols,
                index_col=False)
            # hkl = str(tuple(np.array(df.loc[0,['h','k','l']],dtype=int)))
            hkl = ','.join(df.loc[0,['h','k','l']].values)
            self.df_refl.loc[hkl,cols[:-2]] = [list(df[c]) for c in cols[:-2]]
            self.df_refl.loc[hkl,'frame'] = list(np.array(df['frame'],dtype=int))
            self.df_refl.loc[hkl,'omega'] = list(df['omega'])
            # print(hkl)

    def load_refl(self):
        f = open(self.file_profiles, "r")
        # number of reflection
        n = ([])
        h_list = ([])  # h
        k_list = ([])  # k
        l_list = ([])  # l
        dstar_list = ([])  # reciprocal d-spacing
        s_list = ([])  # deviation parameter, A^-1
        rsg_list = ([])  # the ratio between the excitation error of the reflection
        # and the maximum excitation error spanned by the oscilation of the frame
        # Thus, if |Rsg|=0, the reflection is in the diffraction condition exactly
        # in the middle of the frame. If |Rsg|=1, the reflection is in exact Bragg
        # condition at the edge of the wedge covered by the frame.
        Iobs_list = ([])  # measured intensity
        sigma_list = ([])  # SD, as in I/sigma
        Ifit_list = ([])  # fitted intensity
        frame_list = ([])  # frame number
        frame_max = 0  # max frame number
        omega_list = ([])  # rotation angle
        while(True):
            # read next line
            line = f.readline()
            # empty line is EOF
            if not line:
                break
            if (line.find("#") >= 0):  # we have a new reflection
                frame = ([])  # list of frames for this reflection
                omega = ([])  # list of rotations for this reflection
                s = ([])  # list of deviation parameters for this reflection
                Iobs = ([])  # observed intensities for this reflection
                Ifit = ([])  # fit intensities for this reflection
                rsg = ([])  # rsg, see above
                sigma = ([])  # dunno
                n.append(int(line[line.find("#")+1:]))
                # first line of data for this reflection- never empty, we hope
                line = f.readline()
                h_list.append(int(line[0:4]))
                k_list.append(int(line[4:8]))
                l_list.append(int(line[8:12]))
                dstar_list.append(float(line[12:26]))
                s.append(float(line[26:40]))
                rsg.append(float(line[40:54]))
                Iobs.append(float(line[54:70]))
                sigma.append(float(line[70:84]))
                Ifit.append(float(line[84:96]))
                frame.append(int(line[96:100]))
                omega.append(float(line[100:109]))
                if (int(line[96:100]) > frame_max):
                    frame_max = int(line[96:100])
                # second & subsequent lines of data for this reflection
                line = f.readline()
                while (line[0] != "\n"):
                    s.append(float(line[26:40]))
                    rsg.append(float(line[40:54]))
                    Iobs.append(float(line[54:70]))
                    sigma.append(float(line[70:84]))
                    Ifit.append(float(line[84:96]))
                    frame.append(int(line[96:100]))
                    omega.append(float(line[100:109]))
                    if (int(line[96:100]) > frame_max):
                        frame_max = int(line[96:100])
                    line = f.readline()
                # add each parameter to their arrays
                # frame_list.append([frame])
                # omega_list.append([omega])
                # s_list.append([s])
                # Iobs_list.append([Iobs])
                # Ifit_list.append([Ifit])
                # rsg_list.append([rsg])
                # sigma_list.append([sigma])
                # print(tuple(h_list[-1],k_list[-1],l_list[-1]))
                refl = str((h_list[-1],k_list[-1],l_list[-1]))
                self.df_refl.loc[refl,'dstar']  = dstar_list[-1]
                self.df_refl.loc[refl,'s']      = s
                self.df_refl.loc[refl,'rsg']    = rsg
                self.df_refl.loc[refl,'Iobs']   = Iobs
                self.df_refl.loc[refl,'sigma']  = sigma
                self.df_refl.loc[refl,'Ifit']   = Ifit
                self.df_refl.loc[refl,'frame']  = frame
                self.df_refl.loc[refl,'omega']  = omega

    def load_rock(self):
        # Iobs is ordered according to deviation parameter s
        # reorder according to frame ID, and extract the peak over a frame range 2*ww+1
        # peak_f = ([])  # frame numbers for the extracted range
        # peak_s = ([])  # deviation parameters s for the extracted range
        # peak_I = ([])  # intensities for the extracted range
        # width of the extracted rocking curve +/-
        ww = self.ww
        # n_refl = len(self.h)
        # for i in range(n_refl):
        for i,r in self.df_refl.iterrows():
            # frame list for this reflection
            # c = np.array(sum(r.frame, []))
            c = np.array(r.frame,dtype=int)
            # Iobs for this reflection
            # a = np.array(sum(r.Iobs, []))
            a = np.array(r.Iobs)
            # s for this reflection
            # b = np.array(sum(r.s, []))
            b = np.array(r.s)
            # re-order according to frame number
            # print(a.shape,argsort(c))
            f1 = c[np.argsort(c)]
            a1 = a[np.argsort(c)]
            b1 = b[np.argsort(c)]
            # extract the peak in a 2*ww-frame window around the max
            # the different ways here just limit the window to the actual frame list
            pp = np.argmax(a1)  # pp=peak position in this sub-list
            if(pp-ww < 0):  # peak is close to the start
                ff = f1[:pp+ww]  # frame list
                ss = b1[:pp+ww]  # s list
                II = a1[:pp+ww]  # intensity list
            elif(pp+ww > len(a1)):  # peak is close to the end
                ff = f1[pp-ww:]
                ss = b1[pp-ww:]
                II = a1[pp-ww:]
            else:
                ff = f1[pp-ww:pp+ww]
                ss = b1[pp-ww:pp+ww]
                II = a1[pp-ww:pp+ww]
            # peak_f.append(ff)
            # peak_s.append(ss)
            # peak_I.append(II)
            refl = r.name
            # print(refl,ff,ss,II)
            # self.df_refl.loc[refl,['peak_f','peak_s','peak_I']] = [ff,ss,II]
            self.df_peak.loc[refl,'f'] = ff
            self.df_peak.loc[refl,'s'] = ss
            self.df_peak.loc[refl,'I'] = II


    def import_rocks(self):
        cols = ['dstar','s','rsg','Iobs','sigma','Ifit','frame','omega']
        self.df_refl=pd.DataFrame(columns=cols,dtype=object)
        self.load_refl()
        print('... ordering ...')
        self.df_peak = pd.DataFrame(columns=['f','s','I'],dtype=object,)
        self.load_rock()

    def plot_exp_rock(self,opt='s',show=False):
        outpath = self.paths['rock_exp']
        fi = np.argsort(self.df_int['fmax'])
        # print(fi.values)
        # if not show:opt+='c'
        n = int(np.ceil(np.log10(self.df_int.shape[0])))
        for i,(refl,r) in enumerate(self.df_peak.iloc[fi].iterrows()):
            name = '%s.%s.png' %(str(i).zfill(n),
                refl[1:-1].replace(', ','_'))
            x = np.array(r.f)
            y = np.array(r.I)
            dsp.stddisp([x,y,'b-o'],labs=['Frame','Intensity'],title=refl,
                xylims=[x.min(),x.max()+1,1e2*np.floor(y.min()/1e2),1e2*np.ceil(y.max()/1e2)],
                xyTicks=[5,y.max()/10],
                name = os.path.join(outpath,name) ,opt=opt)
            if show:plt.show()
            plt.close()

    def integrate_exp_intensity(self,thresh = 600,plot_rock=False,save_rock=False):
        f_max = ([])  # frame number for the peak
        I_int_expt_raw = ([])  # raw integrated intensities
        I_int_expt_Lor = ([])  # Lorentz scaled integrated intensities
        ds = ([])  # delta s for the peak
        outpath = self.paths['rock']

        for i,r in self.df_peak.iterrows():
            x = np.array(r.f)
            # default is that background is the minimum value
            back = np.where(r.I >= thresh, min(r.I), r.I)
            # alternative default: background is threshold value
            # back = np.where(peak_I[i]>=thresh, 1500, peak_I[i])
            ItB = back*1.0
            # we assume all data points below thresh are background
            if (min(r.I) < thresh):
                # fit a straight line to the background
                ItB = np.where(r.I >= thresh, np.nan, r.I)
                # squeeze & remove the points above background
                x1 = x[~np.isnan(ItB)]
                b1 = ItB[~np.isnan(ItB)]
                if (len(x1) > 1):
                    # least squares linear fit
                    fit = np.polyfit(x1, b1, 1)
                    # background interpolated under the peak
                    back = x*fit[0]+fit[1]
            # Intensities for the plot
            It0 = np.where(r.I < thresh, np.nan, r.I-back)
            # Intensities for integration (can't have NaNs)
            ItI = np.where(r.I < thresh, 0, r.I-back)
            # Integrated intensity without Lorentz scaling
            I_int_expt_raw.append(sum(ItI))
            # delta s per frame for this reflection
            deltas = abs((r.s[1]-r.s[0]))
            ds.append(deltas)
            # Apply Lorentz scaling, multiply by delta s
            I_int_expt_Lor.append(sum(ItI)*deltas)
            # centroid of thresholded rocking curve
            centr = np.dot(r.f, ItI)/np.sum(ItI)
            # frame number for the peak
            f_max.append(centr)
            # f_max.append(ff[np.argmax(II)])  # simple maximum, deprecated
            if (plot_rock):
                name = r.name
                fig,ax = dsp.create_fig(figsize=(6, 4))
                # ax = fig.add_subplot(111)
                ax.plot(x, r.I, color='grey')
                ax.bar(x, It0, color='green')
                ax.plot(x, back, color='red')
                ax.scatter(x, ItB, s=3, color='red')
                ax.set_xlabel('Frame')
                ax.set_ylabel('Intensity')
                plt.suptitle(name)
                plt.show
                if (save_rock):
                    plotname = os.path.join(outpath, name + '.png')
                    plt.savefig(plotname)
                    plt.close()

        print('integration done')
        self.df_int = pd.DataFrame(
            np.array([f_max,ds,I_int_expt_raw,I_int_expt_Lor]).T,
            columns=['fmax','ds','Iint_raw','Iint_lor'],
            index=self.df_refl.index)

    def correct_rock(self):
       return

    ###########################################################################
    ###########################################################################
    ### Generate and launch simulation
    ###########################################################################
    ###########################################################################
    def write_felix_inp(self,thicks,sig_fig=1e5):
       return


    ###########################################################################
    ###########################################################################
    ### post process : LACBED patterns,simulated integration
    ###########################################################################
    ###########################################################################
    def list_LACBED_hkl(self):
       return

    def import_sims(self,nwx=None):
        cols = ['sim','thick','nwx','refl','file']
        self.df_sims = pd.DataFrame(columns=cols)
        # print()
        sims_dir = glob.glob(os.path.join(self.path,self.xtal,'*/'))
        sims = np.sort([int(s.split('/')[-2]) for s in sims_dir])
        for sim,sim_d in zip(sims,sims_dir):
            folders = glob.glob(os.path.join(sim_d,'*/'))
            for d in folders:
                # print(d.split('/'))#[-1].split('_')[2:])
                thick,nwxs = d.split('/')[-2].split('_')[2:]
                # nwx0 = tuple([int(i) for i in nwxs[:-1].split('x')])
                nwx0 = int(nwxs[:-1].split('x')[0])
                if not type(nwx)==type(None):
                    if not int(nwx0)==nwx:
                        continue
                files = glob.glob(os.path.join(d,'*'))
                refls = [f.split('_')[-1][:-4] for f in files]
                hkl = [(',-'.join(','.join(r.split('+')).split('-'))[1:]) for r in refls]
                n = len(files)
                self.df_sims = pd.concat([
                    self.df_sims,
                    pd.DataFrame(np.array([[sim]*n,[thick]*n,[nwx0]*n,hkl,files]).T,columns=cols),
                    ])
        # print('done')

    def show_lacbed(self,show=0,**kwargs):
        for i,r in self.df_sims.iterrows():
            I = np.reshape(np.fromfile(r.file,dtype=np.float64),(int(r.nwx),)*2)
            # tle = os.path.basename(r.file)
            tle='refl=%s, thick=%s, sim %s,' %(r.sim,'(%s)' %r.refl,r.thick)
            dsp.stddisp(im=[I],pOpt='im',title=tle,
                name=r.file.replace('.bin','.png'),**kwargs)
            if show:plt.show()


    def find_best_row_sim(self):
        ''' Find the best row for each simulation'''
        return
    def extract_sim_integrated_intensity(self):
        return
    def calculate_dynamical_Rfactors(self):
        return



from utils import*
from scattering_factors import wavelength,
figpath=get_figpath(__file__,"/figures/")

def plot_interaction_param(npts=1000,**kwargs):
    KE=np.linspace(10,1000,npts) #keV
    lam = wavelength(KE)
    sigma_e = pi/(lam*KE)                    #rad/A*keV
    sigma_0 = 2*pi*m0*lam*A*eV/h**2*(1+KE/(2*mc2))/(1+KE/mc2)
    sigma = 2*pi*m0*(1+KE/mc2)*lam*A*eV/h**2 #rad/(m*V)
    plts = [[KE,sigma*A*kV,'b',"relativistic"],
            [KE,sigma_0*A*kV,'r--','$\pi/\lambda_0 E_0$']]
    stddisp(plts,labs=['$E_0(keV)$','$\sigma(rad/kV.A)$'],**kwargs)


def test_mulslice_params():
    a0,b0=3.995,5.65
    NxNy=[[5,3],[6,4],[7,5],[8,6]]
    N = np.array([128,256,512])
    KE=200;

    lam = wavelength(KE)
    AB_0 = np.array([max(n[0]*a0,n[1]*b0) for n in NxNy])
    dtheta=lam/AB_0;print("dtheta(mrad)=",dtheta*1e3)
    tmax = (N[:,None]/2).dot(dtheta[None,:])*2/3*1e3

    tmax_book=np.array([53.6,107.1,214.3])
    print(tmax);print(tmax_book)

if __name__ == "__main__" :
    plt.close('all')
    #test_mulslice_params()
    plot_interaction_param(npts=1000,legLoc="upper right",lw=2,
        name=figpath+"scattering_param.svg",opt='s',figsize='f')



from utils import*
datpath=get_figpath(__file__,'/dat/')
figpath=get_figpath(__file__,'/figures/')

# import from slicelib
def import_vatom(files=[],**kwargs):
    if not isinstance(files,list) : files=[files]
    plts,cs=[],getCs('jet',len(files))
    for file,c in zip(files,cs) :
        lab = file.split('_')[0]
        r,v = np.loadtxt(datpath+file).T
        plts+=[[r,v/1e3,c,lab]]
    stddisp(plts,labs=["$r(A)$","$V_a(kV)$"],xylims=[0,0.5,0,20],**kwargs)

def import_vzatom(files=[],labels=[],**kwargs):
    if not isinstance(files,list) : files=[files]
    if not labels : labels = files
    plts,cs=[],getCs('jet',len(files))
    for file,c,lab in zip(files,cs,labels) :
        #lab = file #.split('_')[0]
        r,v = np.loadtxt(datpath+file).T
        plts+=[[r,v,c,lab]]
    stddisp(plts,labs=["$r(A)$","$V_a(V.A)$"],xylims=[0,0.5,0,5000],**kwargs)

##### Import from atompot
def import_structure_factor(files,imopt='',fe=1,**kwargs):
    if not isinstance(files,list) : files=[files]
    #unit cell spacing and atomic form factor
    if not imopt:
        nx = 20
        Xq = np.arange(nx)[:,None].dot(np.array([1,1])[None,:]).T/ax
        Yq = np.tile([0,1],[nx,1]).T
        plts=[[ [0,0],[0,1],[(0.5,1.0,0.5),'--'],'$dq_{unitcell}$']]
        plts+=[[Xq,Yq,[(0.5,1.0,0.5),'--'],'']]
        if fe:
            q0,fq_e = get_elec_atomic_factors(['Si'],q=np.linspace(0,5,1000))
            fe2 = np.abs(fq_e[0]/(ax*by))**2;fe2/=fe2.max() #A^{-1}
            plts+=[[q0,fe2,'g--','$f_e^2(Si)$']]
    cs=getCs('Blues',len(files))
    for c,file in zip(cs,files):
        lab=file.split('.')[0].split('_')[1]
        dqx,dqy=1/np.array(lab.split('x'),dtype=int)/np.array([ax,by])
        re_im = np.loadtxt(datpath+file);
        #keep only half
        N=re_im.shape[1];N2=2*int(N/2)
        re,im = re_im.reshape([2,N,N])
        Fq = re+1J*im
        S = np.abs(Fq)**2;#S/=cz*A
        S/=S.max();S[S<1e-10] = 1e-10
        if imopt :
            stddisp(im=S,imOpt=imopt,xylims=[0,30,0,30],legOpt=0,**kwargs)
        else :
            q,Sq = dqx*np.arange(N2),S[0,:N2].T #np.log10(S[0,:N2].T)
            plts+=[[q,Sq,[c,'*-'],lab]]
        #plts+=[[q,np.real(Fq),'c*-',''],[q,np.imag(Fq),'m*-','']]
    if not imopt:
        stddisp(plts,labs=['$q(A^{-1})$','$I_{h0}$'],xylims=[0,2,0,Sq.max()],
                legLoc='upper right',axPos=[0.2,0.75,0.2,0.75],**kwargs)

if __name__ == "__main__":
    plt.close('all')
    # import_structure_factor(['si110_2x2.txt'],imopt='c',name=figpath+'si110_S_2D.png',opt='s')
    # import_structure_factor(['si110_2x2.txt','si110_10x10.txt'],opt='s',name=figpath+'si110_S_1D.svg',gridOn=0  )
    #     inset={'axpos':[0.2,0.18,0.25,0.25],'xylims':[0,150,0,0.02]})
    # plot_fe(qmax=3,Nx=20,opt='s',name=figpath+'Si_vg.svg',axPos=[0.2,0.125,0.75,0.8],setPos=1,lw=3,gridOn=0)
    import_vatom(datpath+'Si_vatom.txt',opt='s',name=figpath+'Si_va.svg',axPos=[0.2,0.125,0.75,0.8],setPos=1,lw=3)
    # import_vzatom(['Si_vzatom.txt','Si_vzatomLUT.txt'],labels=['$Si_{vz}$','$spline_{fit}$'],opt='p',name=figpath+'Si_vz.svg',axPos=[0.2,0.125,0.75,0.8],setPos=1,lw=3)



import numpy as np
import mrcfile,tifffile
from utils import glob_colors as colors
import os,glob,pickle5

def mrc2tiff(mrc_file,outpath):
    tiff_file = os.path.basename(mrc_file).replace('.mrc','.tiff')
    im = imread(mrc_file).copy()
    im -= im.min()
    im*=50/im.mean()
    # im*=1e4/im.max()
    # im = np.array(im,dtype=np.uint16)
    imwrite(os.path.join(outpath,tiff_file),im)

#### writer
def mrc_writer(filename,im0):
    mrc = mrcfile.open(filename,'r+')
    mrc.set_data(np.array(im0,dtype=type(mrc.data[0,0])))
    mrc.flush()
    mrc.close()

def tiff_writer(tiff_file,im):
    tifffile.imwrite(tiff_file,im)
img_writers = {
    'mrc' : mrc_writer,
    'tiff': tiff_writer,
}
def imwrite(filename,im):
    fmt=filename.split('.')[-1]
    img_writers[fmt](filename,im)
    print(colors.yellow+filename+colors.green+' saved'+colors.black)


#### reader
def tiff_reader(tiff_file)  :
    with open(tiff_file,'rb') as f:
        I =tifffile.imread(f)
    return I
def mrc_reader(mrc_file):
    with mrcfile.open(mrc_file) as mrc:
        return mrc.data
img_readers = {
    'mrc' : mrc_reader,
    'tiff': tiff_reader,
}
fmts = list(img_readers.keys())#['mrc','tiff']
def imread(filename):
    fmt=filename.split('.')[-1]
    return img_readers[fmt](filename)


def strong_beams(dfG,
        tol:float=1e-2,n:int=10,
        opt:str='F'):
    """keep the n strongest beams, i.e. those with intensity I>tol not including central beam

    Parameters
    ----------
    dfG
        the bloch dataframe
    tol
        strong beam criteria
    n
        number of strong beams
    opt
        include Origin(O) Friedel pairs(F)
    Returns
    ----------
    x
        list of beams
    """
    dfM=dfG.copy()
    if not 'O' in opt:dfM = dfM.drop(str((0,0,0)))
    dfM = dfM.sort_values('I',ascending=False)[:2*n]

    df = dfM.loc[dfM.I>tol]
    if not 'F' in opt:
        df=remove_friedel_pairs(df)

    if df.shape[0]>n:df = df[:n]

    # print(dfM)
    return list(df.index.values)

def remove_friedel_pairs(dfM):
    refl=[]
    for h,hkl in zip(dfM.index,dfM[['h','k','l']].values):
        if not str(tuple(-hkl)) in refl:
            refl.append(h)
    print('removing Friedel pairs')
    dfM=dfM.loc[refl]
    return dfM

def get_inp(npx,nbeams,u,keV,thicks,out=None):
    inp = """# Input file for Felix version :
# ------------------------------------

# ------------------------------------

# control flags
IWriteFLAG                = 0
IImageFLAG                = 1
IScatterFactorMethodFLAG  = 0
IBlochMethodFLAG          = 0
IMaskFLAG                 = 1
IHolzFLAG                 = 0
IAbsorbFLAG               = 0
IAnisoDebyeWallerFlag     = 0
IByteSize                 = 8
    """

    inp+="""
# radius of the beam in pixels
IPixelCount               = %d

# beam selection criteria
IMinReflectionPool        = %d
IMinStrongBeams           = %d
IMinWeakBeams             = 0
    """ %(npx,nbeams,nbeams)

    inp+="""
# crystal settings
RDebyeWallerConstant      = 0.0
RAbsorptionPer            = 7.8

# microscope settings
ROuterConvergenceAngle    = 3.03
IIncidentBeamDirection    = [%f,%f,%f]
IXDirection               = [1,1,0]
INormalDirection          = [-1,1,0]
RAcceleratingVoltage (kV) = %.1f
RAcceptanceAngle          = 0.0
    """ %(u[0],u[1],u[2],keV)

    inp+="""
# Image Output Options
RInitialThickness        = %.1f
RFinalThickness          = %.1f
RDeltaThickness          = %.1f
IReflectOut              = 1
    """ %(thicks[0],thicks[1],thicks[2])

    inp+="""
#Refinement Specific Flags
IRefineModeFLAG          = S
IWeightingFLAG           = 0
IRefineMethodFLAG        = 3
ICorrelationFLAG         = 2
IImageProcessingFLAG     = 4
RBlurRadius              = 0.0
INoofUgs                 = 40
IAtomicSites             = (1,2)
IPrint                   = 0
RSimplexLengthScale      = 20.0
RExitCriteria            = 0.00001
    """

    if out:
        with open(out,'w') as f:f.write(inp)
    else:
        return inp

def load_bloch(path='',tag=None,file='',v=1):
    """load a saved Bloch object from
    path : path to simulation folder
    tag : identifier for a bloch simu in case there are several possibilities
        OR
    filename : directly provide the pickle file (.pkl)
    """
    # first tr toget a simu from the tag
    files = [f for f in glob.glob(os.path.join(path,'*.pkl'))]
    if not isinstance(tag,type(None)):
        if isinstance(tag,int):
            file = files[tag]
        else:
            files = [f for f in glob.glob(os.path.join(path,'*.pkl')) if tag in f]
        if len(files)==1:
            file = files[0]
    if file:
        with open(file,'rb') as f : obj = pickle5.load(f)
        if v:print(colors.green+'loaded:' +colors.yellow+file+colors.black);
        return obj
    else:
        if len(files):
            print(colors.red+'mulitple simus available with tag %s:' %tag);
            print(colors.yellow,[os.path.basename(f) for f in files],colors.black)
        else:
            print(colors.red+'no simus available with in "%s" with tag %s' %(path,tag))

def get_fz(opts):
    """
    mapping for function to apply

    Parameters
    ---------
    opts
        r(real) i(imag) a(angle) m(mag) l(log10(|Fhkl|+1)) 2(mag^2) L(logM)
    """
    if not opts:opts='m'
    keys   =['r','i','a','m','l','2','L']
    fs     = [np.real,np.imag,np.angle,np.abs,logF,abs2,logM]
    fs_str = ['real part','imag part','phase','magnitude','$\log_{10}(|F|+1)$','$|F|^2$','$-\log_{10}$']
    fz     = dict(zip(keys,fs))[opts]
    fz_str = dict(zip(keys,fs_str))[opts]
    return fz,fz_str

abs2 = lambda F:np.abs(F)**2
logF = lambda F:np.log10(np.abs(F)+1)
logM = lambda F:-np.log10(np.maximum(F,1e-8))

# def show_beams(self,F='I',fopts='m',opts='xN',mag=500,cutoff=0,cmap='Greens',**kwargs):
#     """Display beam values
#     - F : 'I'(intensity),'S'(scattered beams),'Vg'(potential)
#     - opts  : 'x'(include projection of x axis), 'N'(normalize)
#     - fopts : see get_fz
#     """
#     fz,fz_str = get_fz(fopts)
#     F_str = {'L':'Lattice','Vg':'Potential $V_{g}$',
#         'S' :'Scattered beams Bloch $S_{g}$',
#         'I' :'Intensity Bloch$I_{g}$',
#         'Sg':'Scattered beams Kinematic $S_{g,kin}$',
#         'Ig':'Intensity kinematic $I_{g,kin}$',
#         'Sw':'Excitation error $\zeta_{g}$'}[F]
#
#     tle = '%s, %s, thickness=%d$\AA$'  %(F_str,fz_str,self.thick)
#     px,py = self.df_G[['px','py']].values.T
#     Fvals = fz(self.df_G[F])
#     if 'N' in opts:Fvals/=Fvals.max()
#     # mag /= Fvals.max()
#     scat  = [px,py,Fvals*mag/Fvals.max(),Fvals]
#     plts = [[0,0,'b+','']]
#     if 'x' in opts:plts+=[ [[0,self.e0x],[0,0],'r','']]
#     if not cutoff:cutoff = Fvals.max()
#     # print(Fvals.max())
#     fig,ax = dsp.stddisp(plts,lw=2,scat=scat,caxis=[0,cutoff],cmap=cmap,cs='S',title=tle,**kwargs)



from utils import*
plt.close('all')

rho   = 0.106 #atoms/A^3
sig_e = np.array([0.001,0.0025,0.005]) #A^2
le_s  = 1/(sig_e*rho) #A

z = np.linspace(0,10000,1000) #A
for i,le in zip(range(le_s.size),le_s):
    Pcoh = np.exp(-z/le)
    Pkin = z/le*np.exp(-z/le)
    Pdyn = 1-np.exp(-z/le)*(1+z/le)
    S = Pcoh+Pdyn+Pkin

    tle = r'$\sigma_e=%.3f\AA^2$, $l_e$=%.d nm' %(sig_e[i],int(le/10))
    plts = [[z/10,Pcoh,'r','$coh$'],[z/10,Pkin,'g','$kin$'],[z/10,Pdyn,'b','$dyn$'],[z/10,S,'k']]
    dsp.stddisp(plts,lw=2,labs=[r'$z(nm)$','P(z)'],title=tle,
        name='figures/Pcoh_kin_dyn%d.svg' %i, opt='ps',fonts={'leg':20,'title':20})



import matplotlib.pyplot as plt
import numpy as np
from math import pi,sqrt,gcd
from matplotlib.patches import Rectangle
from crystals import Crystal
from utils.glob_colors import*
import utils.displayStandards as dsp
import utils.handler3D as h3d

cs = {1:unicolor(0.75),3:(1,0,0),6:unicolor(0.25),7:(0,0,1),8:(1,0,0),16:(1,1,0)}

wbs = np.array([1,3,6,7,8,10])*0.01
wobbles = dict(zip([1,3,6,7,8,16],wbs))
Rot = lambda a:np.array([[np.cos(a),0,np.sin(a)],[0,1,0],[-np.sin(a),0,np.cos(a)]])

##########################################################################
# orient crystal along n_u
def orient_crystal(coords,ez=[0,0,1],n_u=[0,0,1],T=True,theta=0):
    ''' Rotate the object so that n_u becomes e_z [1] :\n
    - coords : 3xN (or Nx3 array if T=True)
    - n : axis to align e_z against
    - T : transpose if True (input is Nx3)
    [1] [rotation matrix from axis and angle](https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle)
    '''
    e_z,n = np.array(ez)/np.linalg.norm(ez),np.array(n_u)/np.linalg.norm(n_u)
    n_not_ez = np.linalg.norm(n-e_z)
    if n_not_ez:
        u,ct = np.cross(n,e_z), np.dot(e_z,n)
        u /= np.linalg.norm(u)
        st = np.sqrt(1-ct**2)
        ux,uy,uz = u
        ux2,uy2,uz2 = u**2
        R = np.array([
            [ct+ux2*(1-ct), ux*uy*(1-ct)-uz*st, ux*uz*(1-ct)+uy*st ],
            [uy*ux*(1-ct)+uz*st, ct+uy2*(1-ct), uy*uz*(1-ct)-ux*st ],
            [uz*ux*(1-ct)-uy*st, uz*uy*(1-ct)+ux*st, ct+uz2*(1-ct) ],
            ])
        if T :
            coords = R.dot(coords.T).T
        else:
            coords = R.dot(coords)
    if theta :
        t = np.deg2rad(theta)
        st,ct = np.sin(t),np.cos(t)
        if T :
            coords = np.array([[ct,st,0],[-st,ct,0],[0,0,1]]).dot(coords.T).T
        else:
            coords = np.array([[ct,st,0],[-st,ct,0],[0,0,1]]).dot(coords)
    return coords


def get_crystal_rotation(u=None,omega=0,eo='z',alpha=0):
    '''Return the rotation matrix of alpha about u
    - alpha : float - rotation angle
    - u     : [ux,uy,uz] - rotation axis
    - omega : float - angle of rotation of the rotation axis
    - eo    : 'x','y','z' - normal plane to the rotation for computing u from omega
    '''
    omega,alpha = np.deg2rad(omega),np.deg2rad(alpha)
    if not u:
        if   eo=='x':u = [0,np.cos(omega),np.sin(omega)]
        elif eo=='y':u = [np.cos(omega),0,np.sin(omega)]
        elif eo=='z':u = [np.cos(omega),np.sin(omega),0]
    u = np.array(u)
    ux,uy,uz = u
    ux2,uy2,uz2 = u**2
    ct,st = np.cos(alpha),np.sin(alpha)
    R = np.array([
            [ct+ux2*(1-ct), ux*uy*(1-ct)-uz*st, ux*uz*(1-ct)+uy*st ],
            [uy*ux*(1-ct)+uz*st, ct+uy2*(1-ct), uy*uz*(1-ct)-ux*st ],
            [uz*ux*(1-ct)-uy*st, uz*uy*(1-ct)+ux*st, ct+uz2*(1-ct) ],
            ])
    return R

# rotate crysral for periodic run
def rotate_xyz(file,nx,nz,Nx=40,Nz=5,name='',opt='',**kwargs):
    '''Generates .xyz files corresponding to individual rotation
    - file : path to cif file (.xyz files will be put in the same folder )
    - Nrot : number of rotation to perform within 0,pi/2
    '''
    crys = file
    if isinstance(file,str):
        crys = Crystal.from_cif(file)
        if not name : name = file.replace('.cif','')

    a1,a2,a3    = np.array(crys.lattice_vectors)
    ax1,by2,cz3 = np.array(crys.lattice_parameters[:3])
    nxz   = int(cz3/ax1)
    #int(nx.max()+nxz*nz.max()), int(nx.max()/nxz+nz.max())
    xy0   = (0,(Nz-nz.max())*cz3)
    print('super cell calculation ...')
    crys0   = crys.supercell(Nx,1,Nz)
    pattern = np.array([
        [a.atomic_number]+list(a.coords_cartesian)+[a.occupancy,wobbles[a.atomic_number]] for a in crys0.atoms])
    coords0 = np.array([a.coords_cartesian for a in crys0.atoms])

    files = []
    Cs = [cs[E] for E in pattern[:,0]]
    for i1,i2 in zip(nx,nz):
        # i1,i2 = i,Nrot-i; n0 = gcd(i1,i2)
        # i1,i2 = i1/n0,i2/n0;#print(i1,i2)
        cz = i1*a1+i2*a3; #print(cz)
        ax = np.array([cz[1],-cz[0]]); #print(cz)
        c,angle = np.linalg.norm(cz),np.arctan(cz[0]/cz[2])
        if not i1%i2*nxz:
            n0 = i1/gcd(i1,i2*nxz)
            if not n0%2 : n0/=2
            a = np.linalg.norm([i2*nxz/i1,1])*cz3*n0
        else:
            a = np.linalg.norm([i2*nxz,i1])*cz3
        # a = np.linalg.norm([i2*nxz,i1])*cz3

        print(green+'i1=%d,i2=%d,angle=%.2f' %(i1,i2,angle*180/np.pi)+black)
        rect=Rectangle(xy0,1.0*a,1.0*c,angle=-angle*180/np.pi,edgecolor='k',linewidth=2,alpha=0.1)
        print('finding points ...')
        idc = rect.contains_points(coords0[:,[0,2]],radius=0.001)
        print('done')
        n   = (i1,0,i2)
        xyz = name+'%d%d%d.xyz' %n
        if opt:
            path = dsp.dirname(name)

            x,z = coords0[:,[0,2]].T;
            x0,z0 = (coords0[idc,:])[:,[0,2]].T
            X,Z = np.hstack([x,x0]),np.hstack([z,z0])
            S = np.array([10]*x.size+[30]*x0.size)
            C = Cs+list(np.array(Cs)[idc,:])
            dsp.stddisp(scat=[X,Z,S,C],patches=[rect],equal=1,figsize='12',
                xyTicks=[ax1,cz3],pOpt='tGeX',xylims=[0,40*ax1,-10*cz3,10*cz3],
                opt=opt,name=path+'figures/'+dsp.basename(xyz).replace('xyz','png'))
            plt.show()

        lat_vec = [a,a2[1],c]#np.array([Nrot*(a3+a1),a2,Nrot*(a3-a1)])#
        cc = coords0[idc,:]
        cc[:,2]-=xy0[1]
        coords  = Rot(-angle).dot(cc.T).T
        pattern[idc,1:4] = coords
        # make_xyz(xyz,pattern[idc,:],np.diag(lat_vec),n,fmt='%.4f',dopt='s')
        files+=[xyz] #dsp.basename(xyz)]
    return files


##########################################################################
#unit cell display
##########################################################################
def show_grid(file,opt='',**kwargs):
    with open(file,'r') as f:l=list(map(lambda s:s.strip().split(' '),f.readlines()))
    pattern = np.array(l[2:-1],dtype=float)
    a1,a2,a3 = np.array(l[1],dtype=float)
    #a1,a2,a3,alpha,beta,gamma=crys.lattice_parameters
    if opt:
        Z,x,y,z = pattern[:,:4].T
        Z = np.array(Z,dtype=int)
        C = [cs[E] for E in Z]
        pps = [Rectangle((0,0),a1,a3,linewidth=2,edgecolor='b',alpha=0.1)]

        fig,ax = dsp.stddisp(labs=['$x$','$z$'],patches=pps,scat=[x,z,C],ms=50,
            opt=opt,**kwargs)
    return [a1,a2,a3],pattern
    # a,b,angle=a3,a1,beta
    # lat_vec = lattice.get_lattice_vec('oblique',a=a,b=b,alpha=angle,v=0)
    # lattice.plot_lattice(lat_vec,opts='',nh=3,nk=3,
    #     ax=ax,labs=['$z$','$x$'],lw=2,
    #     pOpt='tG',pad=1,xyTicks=[1.5,5.0],xylims=[-6,15,0,21])#,xyTicksm=1.0,xylims=[0,100,0,100])

def show_unit_cell(ax,n=[0,0,1],ez=[0,0,1],a=0.2,c='b',lw=2,ls='-',lat_params=[1,1,1]):
    '''Orthorombic unit cell from lattice parameters  '''
    Xfaces = get_cubic_cell_mesh(lat_params)
    #change cube orientation and plot
    for X,i in zip(Xfaces,range(len(Xfaces))) :
        x,y,z = X
        coords = np.array([x,y,z]).reshape(3,4).T
        coords = orient_crystal(coords,n_u=n,ez=ez)
        x,y,z  = coords.T.reshape(3,2,2)
        Xfaces[i] = [x,y,z]
        ax.plot_surface(x,y,z,color=c,alpha=a+(i==0)*0.2,linestyle=ls,linewidth=lw,edgecolor=c)
    #cube diagonals
    # (x0,x1),(y0,y1),(z0,z1) = np.array(Xfaces[0])[:,:,0]
    # (x2,x3),(y2,y3),(z2,z3) = np.array(Xfaces[3])[:,:,1]
    # ax.plot([x0,x3],[y0,y3],[z0,z3],'--',color=dsp.unicolor(0.5))
    # ax.plot([x1,x2],[y1,y2],[z1,z2],'--',color=dsp.unicolor(0.5))

def get_cubic_cell_mesh(lat_params,opt=0):
    a1,a2,a3 = lat_params
    #Cube faces
    (x00,y00),z00 = np.meshgrid([0,a1],[0,a2]), np.zeros((2,2))
    (x10,z10),y10 = np.meshgrid([0,a1],[0,a3]), np.zeros((2,2))
    (y20,z20),x20 = np.meshgrid([0,a2],[0,a3]), np.zeros((2,2))
    (x01,y01),z01 = np.meshgrid([0,a1],[0,a2]), np.ones((2,2))*a3
    (x11,z11),y11 = np.meshgrid([0,a1],[0,a3]), np.ones((2,2))*a2
    (y21,z21),x21 = np.meshgrid([0,a2],[0,a3]), np.ones((2,2))*a1
    if opt:
        Xfaces = [  np.array([x00,x10,x20,x01,x11,x21]),
                    np.array([y00,y10,y20,y01,y11,y21]),
                    np.array([z00,z10,z20,z01,z11,z21]),]
    else:
        Xfaces = [[x00,y00,z00],[x10,y10,z10],[x20,y20,z20],
                  [x01,y01,z01],[x11,y11,z11],[x21,y21,z21]]

    return Xfaces



##########################################################################
# misc
##########################################################################
def show_diffraction_planes(r0=1,h=2,npts=10,**kwargs):
    '''Generates the rotation figure'''
    n10,n01,n11,u = [1,0],[0,1],np.array([1,-1])/sqrt(2),[0,0,1]
    Xc0,Yc0,Zc0=get_cylinder(0,pi/2 ,r0,h,npts)
    Xc1,Yc1,Zc1=get_cylinder(pi,3*pi/2,r0,h,npts)
    Xp10,Yp10,Zp10=get_plane(n10,u,w=2*r0,h=h)
    Xp01,Yp01,Zp01=get_plane(n01,u,w=2*r0,h=h)
    Xp11,Yp11,Zp11=get_plane(n11,u,w=2*r0,h=h)
    fig,ax=dsp.stddisp(rc='3d',legOpt=0)
    ax.plot_surface(Xc0, Yc0, Zc0, color='b', alpha=0.2)#,edgecolor='b',linewidth=2)#
    ax.plot_surface(Xc1, Yc1, Zc1, color='b', alpha=0.2 )#,edgecolor='b',linewidth=2)#
    ax.plot_surface(Xp10, Yp10, Zp10, color='b', alpha=0.2,linewidth=2,edgecolor='b'   )#
    ax.plot_surface(Xp01, Yp01, Zp01, color='b', alpha=0.2,linewidth=2,edgecolor='b'   )#
    ax.plot_surface(Xp11, Yp11, Zp11, color='r', alpha=0.4,linewidth=2,edgecolor='r'   )#
    ax.plot([0,0],[0,0],[-h/2,h/2],color ='k',linewidth=2)

    W = 0.75*max(2*r0,h)
    xylims = [-W/2,W/2,-W/2,W/2,-W/2,W/2]
    dsp.standardDisplay(ax,xylims=xylims,axPos=[0,0,1,1],setPos=1,is_3d=1,gridOn=0,ticksOn=0,legOpt=0,**kwargs)

def get_plane(n=[1,0,0],u=[0,1,0],w=1,h=1,x0=[0,0,0]):
    x,y = np.meshgrid([-w/2,w/2],[-h/2,h/2])
    u1,u2 = u,np.cross(n,u)
    Xp = x*u1[0] + y*u2[0] + x0[0]
    Yp = x*u1[1] + y*u2[1] + x0[1]
    Zp = x*u1[2] + y*u2[2] + x0[2]
    return Xp,Yp,Zp

def get_cylinder(ti,tf,r0=1,h=1,npts=10,x0=[0,0,0]):
    t,z,h2 = np.linspace(ti,tf,npts),np.array([-1,1])[:,None],h/2
    Xc = r0*np.tile(np.cos(t),[2,1])  + x0[0]
    Yc = r0*np.tile(np.sin(t),[2,1])  + x0[1]
    Zc = h2*np.tile(z,[1,npts]) + x0[2]
    return Xc,Yc,Zc



########################################################################
# def : tests
########################################################################
def bcc_coords():
    (x,y,z),xc  = np.meshgrid([0,1],[0,1],[0,1]), [0.5,0.5,0.5] #bcc
    coords = np.array([x.flatten(),y.flatten(),z.flatten()]).T
    coords = np.concatenate((coords,[xc]),axis=0)
    return coords

def _test_orient(n=[1,1,1],**kwargs):
    coords = bcc_coords()
    coords = orient_crystal(coords,n)
    scat    = coords.T.tolist()
    fig,ax  = dsp.stddisp(scat=scat,ms=100,rc='3d',opt='',legOpt=0)
    show_unit_cell(ax,n,a=0.1,c='b',lw=2)
    W = np.sqrt(3);xylims = [-W/2,W/2,-W/2,W/2,0,W]
    dsp.standardDisplay(ax,xylims=xylims,gridOn=0,ticksOn=0,legOpt=0,**kwargs)

def _test_trihedron():
    fig,ax = dsp.stddisp(rc='3d',std=0)
    show_trihedron(ax,x0=[0,0,0],cs=None,labs=None,
        lw=1,rc=0.1,h=0.2)
    uvw = np.array([[1,0,0],[0,2,2],[1,1,1]])
    show_trihedron(ax,uvw,x0=[-2,-2,-2],cs=['r','g','b'],labs=['$x$','$y$','$z$'],
        lw=2,rc=0.1,h=0.2)
    dsp.stddisp(ax=ax,pOpt='e',opt='p')

def _test_get_arrow():
    #N = 1
    U = np.array([[0,1,1],[1,1,1],[1,0,1]])
    x0 = [0,0,1]
    cs,lw = dsp.getCs('viridis',U.shape[0]),2
    plots,surfs = [],[]
    for u,cu in zip(U,cs) :
        (x,y,z),(xl,yl,zl) = get_arrow_3d(u,x0,rc=0.1,h=0.2)
        plots += [[xl,yl,zl,cu]]
        surfs += [[x,y,z,cu,None,lw,cu]]

    dsp.stddisp(rc='3d',plots=plots,surfs=surfs,lw=lw,pOpt='e')#,texts=txts)

if __name__ == "__main__":
    plt.close('all')
    #show_diffraction_planes(name='figures/cylinder.png',figopt='t2', opt='p')
    #_test_get_arrow()
    #_test_orient(n=[0,0,1],name='docs_fig/orient_crystal001.png',figopt='2',view=[10,-30],opt='p')#;dsp.crop_fig('docs_fig/orient_crystal001.png',[800,800,600,500])
    #_test_orient(n=[1,1,0],name='docs_fig/orient_crystal110.png',figopt='2',view=[10,-30],opt='p')#;dsp.crop_fig('docs_fig/orient_crystal110.png',[800,800,300,400])
    #_test_orient(n=[1,1,1],name='docs_fig/orient_crystal111.png',figopt='2',view=[10,-30],opt='p')#;dsp.crop_fig('docs_fig/orient_crystal111.png',[800,800,450,350])
    _test_trihedron()



'''Some structure factors in 2D'''
from utils import*
from mpl_toolkits.mplot3d import Axes3D
from scattering_factors import*
import scipy
from scipy.interpolate import interp1d
from scipy.signal.windows import tukey
import utils.FourierUtils as fu
import wallpp.plane_group as pg
figpath=get_figpath(__file__,"/figures/")


def get_potential(fxy,a,npts=2**8,opt='p',**kwargs):
    ixy=np.arange(0,npts)*a/npts
    x,y=np.meshgrid(ixy,ixy)
    f=fxy(x,y)
    F=np.fft.fftshift(np.fft.fft2(f))
    if 'p' in opt:
        xy,kxy=fu.get_extents(a,npts)
        fu.plot_fft2d(f,F,xy,kxy,**kwargs)
    return f,F

def test_get_potential():
    k0 = 2*pi/a*np.array([1.0,0.0]); kx0,ky0=k0
    V = lambda x,y:np.cos(kx0*x+ky0*y)
    f,F = get_potential(V,a,npts=512,opt='p',figsize='12',pad=0.5)
def test_pattern_potential():
    pattern = np.array([[0.25,0.25,2], [0.5,0.5,3]])
    crystal=pg.Wallpaper('p2',a,pattern=pattern,gen=True,ndeg=npts,nh=1,nk=1,pOpt='AVp')
    x,f=crystal.get_potential_unit_cell()
    #F=np.fft.fftshift(np.fft.fft2(f))
    #xy,kxy=fu.get_extents(a,npts)
    #fu.plot_fft2d(f,F,xy,kxy,figsize='12',pad=0.5)

def test_Vx_1D(opt='R'):
    ''' opt :
        Q(plot q,Vq) w(show window function),
        R(plot r,Vr) i(show imaginary)'''
    npts,Npad = 1000,10**4
    elts,qmax = ['C'],6
    q,fq = get_elec_atomic_factors(elts,qmax=qmax,npts=npts)
    fq = fq[0]
    w = tukey(2*npts,alpha=0.5,sym=False)[npts:]
    fqw = w*fq

    #qpad,fqpad = q,fq
    qpad,fqpad = np.linspace(0,int(qmax*Npad/npts),Npad),np.zeros((Npad))
    fqpad[:npts] = fqw
    qpad,fqpad = np.concatenate((-np.flip(qpad),qpad)),np.concatenate((np.flip(fqpad),fqpad))
    if 'Q' in opt:
        plts = [[qpad,fqpad,'b','$fq_{pad}$',2]]
        if 'w' in opt : plts += [[q,fq,'b','$fq$',2],[q,fqw,'r','$fq_w$',2],[q,w,'r--','$w$',1]]
        stddisp(plts,labs=['$q(A^{-1})$','$Fq(A)$'],opt='p')

    ##perform 1D fft
    if 'R' in opt:
        r,Vr = fu.get_iFFT(qpad,fqpad)
        plts = [[r,np.abs(Vr),'g','$V$']]
        if 'i' in opt : plts += [[r,Vr.real,'b','$Re$'],[r,Vr.imag,'r','$Im$']]
        stddisp(plts,labs=['r(A)','$V$'],opt='p',lw=2)



def test_Vx_2D(KE=100,tmax=0.2,npts=100,Npad=1000,opt='QRwp'):
    '''
KE          : incident energy (keV)
tmax        : maximum scattering angle (rad)
npts,Npad   : nb points for normal and padded regions
opt         : Q(plot q,Vq) R(plot r,Vr) w(apply window) p(apply padding)'
    '''
    lam = wavelength(KE)
    ikxy = np.linspace(-1,1,npts)
    qmax = tmax/lam
    qx,qy = np.array(np.meshgrid(ikxy,ikxy))*qmax
    q = np.sqrt(qx**2+qy**2)
    fq = get_elec_atomic_factors(['C'],q)[1][0]
    fq[q>qmax]=0

    if 'w' in opt:
        wf = interp1d(ikxy*qmax,tukey(npts,alpha=0.5))
        w = np.zeros(q.shape);w[q<=qmax]=wf(q[q<=qmax])
        fqw = fq*w
        fq = fqw
    if 'p' in opt:
        qM = qmax*Npad/npts
        iqxy = np.linspace(-qM,qM,Npad)
        (qxpad,qypad),fqpad = np.meshgrid(iqxy,iqxy),np.zeros((Npad,Npad))
        qpad = np.sqrt(qxpad**2+qypad**2)
        fqpad[(np.abs(qxpad)<=qmax) & (np.abs(qypad)<=qmax)] = fq[(qx<=qmax) & (qy<=qmax)]
        qx,qy,fq,npts=qxpad,qypad,fqpad,Npad

    if 'Q' in opt:
        stddisp(im=[qx,qy,fq],labs=['$q_x$','$q_y$'],imOpt='c',pOpt='t')

    if 'R' in opt:
        Vr=np.fft.fftshift(np.fft.ifft2(fq))
        dq = qx[0,1]-qx[0,0];#print(dq)
        r = np.fft.fftshift(np.fft.fftfreq(npts,dq))
        x,y = np.meshgrid(r,r)
        im=[x,y,np.abs(Vr)];print('im_max/re_max=%.2f' %(Vr.imag.max()/Vr.real.max()))
        stddisp(im=im,labs=['$x$','$y$'],imOpt='c',pOpt='t')

    print('tmax=%.2f deg\nqmax=%.2f A^-1\nmax_res=%.2f A' %(tmax*180/pi,qmax,lam/tmax))


#if __name__ == "__main__" :
  #test_get_potential()
  #test_pattern_potential()
  #test_Vx_1D(opt='RQ')
  #test_Vx_2D(npts=50,Npad=200,opt='QRwp',KE=100,tmax=0.2)
